# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Module for miscellaneous and general utilities.
"""

from .exceptions import *
from .provenance import ObjectOrigin, FileOrigin, Transformation, Formatting



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
***********************
Module for Selection
***********************
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from ..utils.py23 import *
from ..utils.dependencies import softimport

numpy = softimport("numpy")
pyparsing = softimport("pyparsing")

# -----------------------------------------------------------------------------
# Selection class
# -----------------------------------------------------------------------------


class ParseResultsWrapper(object):

    def __init__(self, result):
        if not isinstance(result, pyparsing.ParseResults):
            raise TypeError("pyparsing.ParseResults expected!")
        attrs_list = ["lhs", "rhs", "operator", "__getitem__", "asList"]
        if not all(hasattr(result, attr) for attr in attrs_list):
            msg = "pyparsing.ParseResults expected with the following"
            msg += " attributes {0} expected!".format(attrs_list)
            raise TypeError(msg)
        self._result = result

    @property
    def lhs(self):
        ret = self._result.lhs
        if isinstance(ret, pyparsing.ParseResults) and len(ret) == 1:
            return ret[0]
        return ret

    @property
    def rhs(self):
        ret = self._result.rhs
        if isinstance(ret, pyparsing.ParseResults) and len(ret) == 1:
            return ret[0]
        return ret

    @property
    def operator(self):
        ret = self._result.operator
        if isinstance(ret, pyparsing.ParseResults) and len(ret) == 1:
            return ret[0]
        return ret

    def __getitem__(self, i):
        return self._result[i]

    def asList(self):
        return self._result.asList()


class Selection(object):
    """
    Class for dataset selections.
    """

    def __init__(self, selection=""):
        """Default constructor for Selection.

        Parameters
        ----------
        selection: a string.
        """

        selection = selection.replace("&&", "&")
        selection = selection.replace("||", "|")
        self._selection = selection

    @property
    def parsed(self):
        """
        Return the parsed selection
        """

        Optional           = pyparsing.Optional
        Suppress           = pyparsing.Suppress
        Literal            = pyparsing.Literal
        Word               = pyparsing.Word
        nums               = pyparsing.nums
        alphas             = pyparsing.alphas
        oneOf              = pyparsing.oneOf
        OneOrMore          = pyparsing.OneOrMore
        Group              = pyparsing.Group
        opAssoc            = pyparsing.opAssoc
        operatorPrecedence = pyparsing.operatorPrecedence

        openpar    = Literal("(")
        closepar   = Literal(")")
        comma      = Literal(",")
        expop      = Literal('^')
        signop     = Literal('-')
        multop     = oneOf('* / **')
        plusop     = oneOf('+ -')

        functions = Word(alphas)

        number     = Word(nums+".")
        variable   = Word(alphas+nums+"."+"-"+"_")
        operations = operatorPrecedence(number | variable,
                                        [("^", 2, opAssoc.RIGHT),
                                         (signop, 1, opAssoc.RIGHT),
                                         (multop, 2, opAssoc.LEFT),
                                         (plusop, 2, opAssoc.LEFT), ])

        function = Group(functions + Suppress(openpar)
                         + Group(OneOrMore((operations | variable)
                                           + Suppress(Optional(comma))))
                         + Suppress(closepar))

        identifier = function | operations | number | variable

        operator = oneOf("> < >= <= != ==")
        subexpr = identifier.setResultsName("lhs")
        subexpr += operator.setResultsName("operator")
        subexpr += identifier.setResultsName("rhs")

        _and = Literal("&")
        _or = Literal("|")

        exp = operatorPrecedence(Group(subexpr), [(_and, 2, opAssoc.LEFT),
                                                  (_or, 2, opAssoc.LEFT), ])

        ret = exp.parseString(self._selection)[0]
        ret = ParseResultsWrapper(ret)

        return ret

    def evaluateselection(self, dataset, operators, functions):
        """
        Evaluate the selection for a given dataset, taking operators, and
        functions according to the type of the dataset.
        """

        def operation(lhs=None, operator=None, rhs=None):

            if lhs is None and operator == "-":
                return -rhs
            else:
                return operators[operator](lhs, rhs)

        def operand(_operand):
            # Return instance of the operand

            if len(_operand) == 1 and not isinstance(_operand, str):
                _operand = _operand[0]
            elif len(_operand) > 1 and isinstance(_operand,
                                                  pyparsing.ParseResults):

                # treatment of min/max as operand
                if any(m in _operand.asList() for m in list(functions.keys())):
                    func = functions[_operand[0]]

                    args = []
                    for a in _operand[1]:
                        args.append(operand(a))
                    return func(*args)

                # treatment with arithmetical operations as operand
                if any(op in _operand.asList() for op in ["-", "+", "*", "/",
                                                          "^", "**"]):
                    if len(_operand) == 2:
                        return operation(operator=_operand[0],
                                         rhs=operand(_operand[1]))
                    elif len(_operand) > 2:
                        return operation(lhs=operand(_operand[0]),
                                         operator=_operand[1],
                                         rhs=operand(_operand[2:]))

            if isinstance(_operand, str):

                if _operand == "True":
                    return True
                elif _operand == "False":
                    return False
                try:
                    float(_operand)
                except ValueError:
                    return dataset[_operand]
                else:
                    return float(_operand)

            if isinstance(_operand, pyparsing.ParseResults):
                return operand(_operand)

        def loopselection(parsedsel):
            # loop inside the selection, to take into account all subterms,
            # and evaluate the selection from right to left.

            suboperations = []
            bitwiseoperators = []

            for i, s in enumerate(parsedsel):
                if i % 2 == 0:
                    if "&" in s.asList() or "|" in s.asList():
                        suboperations.append(loopselection(s))
                    else:
                        lhs = operand(s.lhs)
                        rhs = operand(s.rhs)
                        suboperations.append(operation(lhs, s.operator, rhs))
                else:
                    bitwiseoperators.append(s)

            while len(bitwiseoperators) > 0:
                loopedselection = operation(suboperations[-2],
                                            bitwiseoperators[-1],
                                            suboperations[-1])
                bitwiseoperators.pop()
                suboperations.pop()
                suboperations.pop()

                suboperations.append(loopedselection)

            return loopedselection

        if "&" in self.parsed.asList() or "|" in self.parsed.asList():
            evaluatedselection = loopselection(self.parsed)
        else:
            s = self.parsed
            lhs = operand(s.lhs)
            rhs = operand(s.rhs)
            evaluatedselection = operation(lhs, s.operator, rhs)

        return evaluatedselection

    @property
    def numpyselection(self):
        """
        Return the selection readable for NumpyDataset
        """

        operators = {">": numpy.greater, "<": numpy.less,
                     ">=": numpy.greater_equal, "<=": numpy.less_equal,
                     "==": numpy.equal, "!=": numpy.not_equal,
                     "&": numpy.bitwise_and, "|": numpy.bitwise_or,
                     "+": numpy.add, "-": numpy.subtract, "/": numpy.divide,
                     "*": numpy.multiply, "^": numpy.power, "**": numpy.power}

        def evaluateufunc(ufunc, *args):
            if len(args) == 2:
                return ufunc(*args)
            else:
                return ufunc.reduce(args)

        def _min(*args):
            ufunc = numpy.minimum
            return evaluateufunc(ufunc, *args)

        def _max(*args):
            ufunc = numpy.minimum
            return evaluateufunc(ufunc, *args)

        functions = {"min": _min, "max": _max, "sin": numpy.sin,
                     "cos": numpy.cos, "tan": numpy.cos,
                     "arcsin": numpy.arcsin,
                     "arccos": numpy.arccos, "arctan": numpy.arctan,
                     "arctan2": numpy.arctan2, "sinh": numpy.sinh,
                     "cosh": numpy.cosh, "tanh": numpy.tanh,
                     "arcsinh": numpy.arcsinh, "arcosh": numpy.arccosh,
                     "log": numpy.log, "log10": numpy.log10,
                     "log1p": numpy.log1p, "exp": numpy.exp,
                     "expm1": numpy.expm1, "sqrt": numpy.sqrt,
                     "abs": numpy.absolute}

        def selection(numpydataset):
            # selection as a function of the numpy dataset
            return self.evaluateselection(numpydataset, operators, functions)

        return selection

    @property
    def pandasselection(self):
        """
        Return the selection readable for possible PandasDataset.
        """

        raise NotImplementedError

    @property
    def rootselection(self):
        """
        Return the selection readable for RootDataset.

        The use the formulate package is consideed in the future
        for better conversion between numexpr and ROOT expression styles.
        """

        selection = self._selection.replace(" & ", " && ")
        selection = selection.replace(" | ", " || ")
        return selection

    def __repr__(self):
        return self._selection

    def __and__(self, other):
        """
        Return new selection = self and other
        """

        if not isinstance(other, Selection):
            return NotImplemented

        if "|" in self._selection:
            lhs = "(" + self._selection + ")"
        else:
            lhs = self._selection

        if "|" in other._selection:
            rhs = "(" + other._selection + ")"
        else:
            rhs = other._selection

        selection = lhs + " & " + rhs

        return Selection(selection)

    def __or__(self, other):
        """
        Return new selection = self or other
        """

        if not isinstance(other, Selection):
            return NotImplemented

        if "&" in self._selection:
            lhs = "(" + self._selection + ")"
        else:
            lhs = self._selection

        if "&" in other._selection:
            rhs = "(" + other._selection + ")"
        else:
            rhs = other._selection

        selection = lhs + " | " + rhs

        return Selection(selection)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
# Licensed under a 3-clause BSD style license, see LICENSE.
# =============================================================================
"""
Set of utilities for comparison of floating point numbers inspired by
- https://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/
and the previous version:
- http://www.cygnus-software.com/papers/comparingfloats/Obsolete%20comparing%20floating%20point%20numbers.htm
"""

__all__ = (
    'distance'      , ## distance in ULPs
    'isclose'       , ## close in  terms of ULPs ?
    'isequal'       , ## isequal using  ULPs and SCALE
    'next_double'   , ## get next double
    )

def distance ( a ,  b ) :
    """
    Distance in ULPS between two (floating point) numbers.
    It is assumed here that  size(long)==size(double) for underlying C-library!

    Example
    -------
    >>> a = ...
    >>> b = ...
    >>> print distance ( a , b )
    """
    if   a == b : return 0
    elif a >  b : return -distance (  b ,  a )
    elif b <= 0 : return  distance ( -b , -a )
    elif a <  0 :
        return distance ( 0 , -a ) + distance ( 0 , b )

    ## here a and b have same sign

    import ctypes

    a , b =  abs ( a ) , abs ( b )

    aa = ctypes.c_double ( float ( a ) )
    bb = ctypes.c_double ( float ( b ) )

    al = ctypes.c_long.from_buffer ( aa ).value
    bl = ctypes.c_long.from_buffer ( bb ).value

    return bl - al


def next_double ( a , ulps = 1 ) :
    """
    Get the ``next-double'' by certain ULPs distance.
    It is assumed here that size(long)==size(double) for underlying C-library!

    Examples
    --------
    >>> a = next_double ( 1 , 1 )
    >>> print a , a - sys.float_info.epsilon
    """

    a = float ( a )
    if   0 == ulps : return  a
    elif a < 0     : return -next_double ( -a , -ulps )
    elif 0 > ulps  :
        d =  distance ( a , 0.0 ) + ulps
        if d < 0 : return -next_double ( 0.0 , -d )

    ## aply a trick  with casting

    import ctypes

    aa  = ctypes.c_double ( a         )
    al  = ctypes.c_long.from_buffer   ( aa ).value
    al  = ctypes.c_long   ( al + ulps )
    aa  = ctypes.c_double.from_buffer ( al ).value

    return aa


def isclose ( a  , b , ulps = 1000 ) :
    """
    Are two floating point numbers close enough (in units is ULPs)?

    Examples
    --------
    >>> a = ...
    >>> b = ...
    >>> print isclose ( a , b , 1000 )
    """
    return ( a == b ) or ulps >= abs ( distance ( a , b ) )


def isequal ( a , b , scale = 1.0 , absdiff = 0.0 , ulps = 1000 ) :
    """
    Are  two numbers ``a'' and ``b''  close enough ?
    Numbers are considered to be equal is :
    - (for absdiff >0) the absolute difference is smaller than absdiff
    - *OR* they are close enough in ULPs
    - *OR* (for scale != 0) the difference is small enough compared to the scale

    Examples
    --------
    >>> isequal(1,1+sys.float_info.epsilon)
    True
    >>> isequal(1,1+sys.float_info.epsilon, ulps = 0 )
    False
    >>> isequal(1,1+sys.float_info.epsilon, scale = 100 , ulps = 0 )
    True
    """
    return ( a == b ) or ( 0 < absdiff and abs ( a - b ) < absdiff ) or isclose ( a , b , ulps ) or ( scale and isclose ( ( a - b ) + scale , scale , ulps ) )



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.dataset.selection module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest
import os

import numpy as np

from skhep.utils.provenance import ObjectOrigin  # these are not imported automatically
from skhep.dataset.numpydataset import *
from skhep.dataset.selection import Selection


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------

def test_constructors():	
	s = Selection()
	
def test_methods():	
	s1 = Selection("x > 1")
	assert s1.__repr__() == "x > 1"
	assert s1.parsed.lhs == "x"
	assert s1.parsed.operator == ">"
	assert s1.parsed.rhs == "1"
	s2 = Selection("y^2 < 2")
	assert s2.__repr__() == "y^2 < 2"
	assert s2.parsed.lhs.asList() == ['y', '^', '2']
	assert s2.parsed.operator == "<"
	assert s2.parsed.rhs == "2"
	s3 = s1 | s2
	assert s3.__repr__() == "x > 1 | y^2 < 2"
	s1 & s3
	assert s3.parsed[0].asList() == s1.parsed.asList()
	assert s3.parsed[1] == "|"
	assert s3.parsed[2].asList() == s2.parsed.asList()
	s4 = s1 & s2
	assert s4.__repr__() == "x > 1 & y^2 < 2"
	assert s4.parsed[1] == "&"
	s5 = Selection("z <= x")
	assert s5.__repr__() == "z <= x"
	assert s5.parsed.lhs == "z"
	assert s5.parsed.operator == "<="
	assert s5.parsed.rhs == "x"
	s6 = s4 | s5
	s6 = s5 | s4
	assert s6.__repr__() == "z <= x | (x > 1 & y^2 < 2)"
	s7 = s5 & s4
	assert s7.__repr__() == "z <= x & x > 1 & y^2 < 2"
	s8 = Selection("-x < 1")
	s9 = ( s4 | s5 ) & s6

	
	


# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Subpackage for the description of datasets.
"""

#-----------------------------------------------------------------------------
# Import statements
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from .numpydataset import NumpyDataset
from .rootdataset import ROOTDataset



# Licensed under a 3-clause BSD style license, see LICENSE.
import numpy as np
import pytest


def pytest_addoption(parser):
    parser.addoption("--cmdopt", action="store", default="test", help="option: 'test' or \
    'generate'. Only use 'generate' if you've changed the tests and need to update the expected \
    output!")


@pytest.fixture
def cmdopt(request):
    return request.config.getoption("--cmdopt")


@pytest.fixture(scope="session")
def data_gen():
    np.random.seed(111)
    data1 = np.random.normal(size=1000)
    data2 = np.random.normal(2, 1, size=1000)
    weights = np.random.uniform(1, 2, size=1000)
    return data1, data2, weights



import os
import numpy as np
import skhep
from skhep.modeling import bayesian_blocks

answer_dir = os.path.dirname(skhep.__file__)+'/../tests/data/modeling'


def test_bayesian_blocks(cmdopt, data_gen):

    be1 = bayesian_blocks(data_gen[0], p0=0.05)
    be2 = bayesian_blocks(data_gen[0], gamma=0.1)
    be3 = bayesian_blocks(data_gen[0], weights=data_gen[2])

    if cmdopt == "generate":
        with open(answer_dir+'/answers_bayesian_blocks.npz', 'wb') as f:
            np.savez(f, be1=be1, be2=be2, be3=be3)
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_bayesian_blocks.npz')
        assert(np.all(be1 == answers['be1']))
        assert(np.all(be2 == answers['be2']))
        assert(np.all(be3 == answers['be3']))
        # assert(np.all(output[1] == answers['be']))



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.utils.provenance module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest

from skhep.utils import *
from skhep.utils.provenance import Provenance, Origin, MultiProvenance  # these are not imported automatically
from skhep.utils.exceptions import *


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------

def test_base_classes():
    with pytest.raises(TypeError):
        Provenance.__init__()
    with pytest.raises(TypeError):
        Origin.__init__()
    with pytest.raises(SkhepTypeError):
        Provenance()
    with pytest.raises(SkhepTypeError):
        Origin()

def test_ObjectOrigin():
    with pytest.raises(TypeError):
        ObjectOrigin()
    with pytest.raises(AssertionError):
        ObjectOrigin([1,2,3])
    prov = ObjectOrigin('array_of_ints')
    assert prov.__repr__() == '<ObjectOrigin>'
    assert prov.detail == 'array_of_ints'

def test_FileOrigin():
    with pytest.raises(TypeError):
        FileOrigin.__init__()
    prov1 = FileOrigin('file.root')
    assert prov1.__repr__() == '<FileOrigin (1 file)>'
    assert prov1.detail == '"file.root"'
    prov1bis = FileOrigin(['file.root'])
    assert prov1bis.__repr__() == '<FileOrigin (1 file)>'
    assert prov1bis.detail == '"file.root"'
    prov3 = FileOrigin(['file1.root', 'file2.root','file3.root'])
    assert prov3.__repr__() == '<FileOrigin (3 files)>'
    assert prov3.detail == '"file1.root","file2.root","file3.root"'
    with pytest.raises(AssertionError):
        prov4 = FileOrigin([1,2,3])

def test_Transformation():
    with pytest.raises(TypeError):
        Transformation.__init__()
    transf = Transformation('all elms * 2')
    assert transf.__repr__() == '<Transformation(all elms * 2)>'
    assert transf.detail == 'all elms * 2'
    transf1 = Transformation('all elms * 2', "detail1", "detail2", transf)
    transf1.detail

def test_Formatting():
    with pytest.raises(TypeError):
        Formatting.__init__()
    forma = Formatting('ROOTDataset', 'DecayTree')
    assert forma.__repr__() == '<Formatting to ROOTDataset(DecayTree)>'
    assert forma.detail == 'ROOTDataset(DecayTree)'
    forma1 = Formatting('ROOTDataset', 'DecayTree', "detail1", forma)
    
def test_MultiProvenance():
    with pytest.raises(TypeError):
        MultiProvenance.__init__()
    with pytest.raises(ValueError):
        MultiProvenance("provenance")
    with pytest.raises(ValueError):
        MultiProvenance.__iadd__(MultiProvenance(),"provenance")
        
    prov1 = FileOrigin('file.root')
    multip = MultiProvenance(prov1)
    assert multip[0].__repr__() == '<FileOrigin (1 file)>'
    assert multip[0].detail == '"file.root"'
    assert multip.detail == '"file.root"'
    transf = Transformation('all elms * 2')
    multip += transf
    assert multip[1].__repr__() == '<Transformation(all elms * 2)>'
    assert multip[1].detail == 'all elms * 2'
    forma = Formatting('ROOTDataset', 'DecayTree')
    multip1 = multip.copy() + MultiProvenance(forma)
    assert multip1[2].__repr__() == '<Formatting to ROOTDataset(DecayTree)>'
    assert multip1[2].detail == 'ROOTDataset(DecayTree)'
    assert multip1.__repr__() == "0: {0}\n1: {1}\n2: {2}".format(prov1, transf, forma)
    multip2 = MultiProvenance(multip1)


# Licensed under a 3-clause BSD style license, see LICENSE.
"""
***********************
Module for NumpyDataset
***********************

The ``NumpyDataset`` class is the implementation of the ``Dataset`` abstract base class
for the [NumPy]_ package.

Note: usage of course requires that NumPy is installed.

**References**

.. [NumPy] http://www.numpy.org/.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import, division

import sys
from types import MethodType

from ..utils.py23 import *
from ..utils.decorators import inheritdoc
from ..utils.dependencies import softimport
from ..utils.provenance import FileOrigin, ObjectOrigin, Transformation, Formatting, MultiProvenance, Provenance
from ..dataset.selection import Selection

from .defs import *

numpy = softimport("numpy")
root_numpy = softimport("root_numpy")


# -----------------------------------------------------------------------------
# NumpyDataset
# -----------------------------------------------------------------------------
class NumpyDataset(FromFiles, ToFiles, NewROOT, Dataset):
    def __init__(self, data, provenance=None, **options):
        """Default constructor for NumpyDataset.

        Parameters
        ----------
        data: a dictionary of equal-length, one-dimensional Numpy arrays or, equivalently, a Numpy record array.
        provenance: history of the data before being wrapped as a NumpyDataset.
        options: none.
        """

        assert self.isrecarray(data) or self.isdictof1d(data)
        self._data = data

        if isinstance(data, dict):
            self.dicttoarray()

        for var in self.variables:
            self.__add_var(var)

        if isinstance(provenance, MultiProvenance):
            self._provenance = provenance.copy()
        else:
            if provenance is None:
                provenance = ObjectOrigin(repr(data))
            if not isinstance(provenance, ( list, tuple ) ):
                provenance = [provenance]

            self._provenance = MultiProvenance(*provenance)

    def copy(self):
        """Get a copy of the NumpyDataset."""
        return NumpyDataset( self.data, self._provenance )

    def select(self, selection = None):
        """
        Apply a selection to the NumpyDataset.

        Parameters
        ----------
        selection: Selection or str

        Returns
        -------
        new NumpyDataset after selection
        """

        if isinstance(selection, numpy.ndarray) and selection.dtype == bool:
          return self.__getitem__(selection)
        elif isinstance(selection, Selection) or isinstance(selection, str):
          if isinstance(selection, str):
            selection = Selection(selection)

          data = self.data[ selection.numpyselection(self) ]

          nbefore    = self.nevents
          nafter     = len(data)
          efficiency = nafter / nbefore
          error      = (( efficiency * ( 1. - efficiency ) ) / nbefore ) ** 0.5

          tr_name    = "Selection, {0}, applied".format(selection)
          tr_detail1 = "#events before = {0}".format(nbefore)
          tr_detail2 = "#events after = {0}".format(nafter)
          tr_detail3 = "Efficiency = {0:.4f} +/- {1:.4f} %".format(efficiency * 100, error * 100)

          provenance = self._provenance + Transformation(tr_name, tr_detail1, tr_detail2, tr_detail3)
          return NumpyDataset(data, provenance)

        else:
          raise ValueError("selection input must be of type 'str', 'Selection', or an Array of booleans not {0}".format(type(selection)))

    @property
    def nevents(self):
        """Get the number of events in the NumpyDataset."""
        return self.__len__()

    @property
    def nentries(self):
        """Get the number of entries in the NumpyDataset. Same as 'nevents'"""
        return self.__len__()

    @property
    def variables(self):
      """
      Get the list of variables in the NumpyDataset, i.e. the content of 'numpy.dtype.names'
      of the stored NumPy array.
      """
      return [var for var in self.data.dtype.names]

    def keys(self):
      """
      Get the list of keys in the NumpyDataset (same as 'variables').
      """
      return [var for var in self.data.dtype.names]

    @staticmethod
    def isrecarray(data):
        is_valid_recarray = isinstance(data, numpy.recarray)
        is_valid_array = (isinstance(data, numpy.ndarray) and
                          isinstance(data.dtype.names, tuple) and
                          len(data.dtype.names) > 0)
        return is_valid_recarray or is_valid_array

    @staticmethod
    def isdictof1d(data):
        is_dict = isinstance(data, dict)
        is_non_empty = len(data) > 0
        has_only_valid_columns = all(isinstance(column, numpy.ndarray) and len(column.shape) == 1
                                     for column in data.values())
        columns_have_valid_shape = all(column.shape == head(data.values()).shape for column in data.values())
        return is_dict and is_non_empty and has_only_valid_columns and columns_have_valid_shape

    def dicttoarray(self):
        """Convert a dictionnary into a structured array.
        If using Python3, byte keys are decoded into string.
        """
        if self.isdictof1d(self.data ) and not self.isrecarray(self.data ):
            if sys.version_info[0] > 2 and any(isinstance(k, bytes) for k in self.data.keys()):
              data = {}
              for k in list(self.data.keys()):
                if isinstance(k, bytes):
                  data[k.decode()] = self.data[k]
                else:
                  data[k] = self.data[k]
            else:
              data = self.data

            dtypes = {'names': list(data.keys()), 'formats': [numpy.dtype(self.data [k].dtype) for k in self.data .keys()]}
            shape  = (len(list(self.data .values())[0]),)
            array  = numpy.zeros(shape,dtypes)

            for k in data.keys():
                array[k] = data[k]

            self._data = array

    @property
    @inheritdoc(Dataset)
    def datashape(self):
        raise NotImplementedError  # TODO!

    @property
    @inheritdoc(Dataset)
    def immutable(self):
        return False

    @property
    @inheritdoc(Dataset)
    def persistent(self):
        return False

    @staticmethod
    def from_file(files, **options):
        """Load a dataset from a file or collection of files.

        Recognizes zipped Numpy (.npz) format.

        Parameters
        ----------
        files: a string file name (glob pattern), iterable of string file names, or an iterable of files open for reading (binary).
        options:
            columns: a set of columns to select from the files.
        """
        requested_columns = options.get("columns")

        columns = None
        data = None

        for x in NumpyDataset._openFileNames(files):
            npzfile = numpy.load(x)

            if requested_columns is not None:
                assert set(requested_columns).issubset(set(npzfile.keys()))

            if columns is None:
                if requested_columns is None:
                    columns = set(npzfile.keys())
                else:
                    columns = set(requested_columns)

            # always read column-wise: collection of 1d arrays in a zip file (.npz)
            # (not too different from what ROOT is, actually)
            newdata = dict((column, npzfile[column]) for column in columns)
            assert NumpyDataset.isdictof1d(newdata)

            if data is None:
                data = newdata
            else:
                data = dict((column, numpy.hstack(
                    data[column], newdata[column])) for column in columns)

        if data is None:
            raise IOError("Empty set of files: {0}".format(files))

        return NumpyDataset(data, FileOrigin(files))

    @inheritdoc(ToFiles, gap="\n")
    def to_file(self, base, **options):
        """

        options: none
        """
        # always write column-wise: collection of 1d arrays in a zip file (.npz)
        # (not too different from what ROOT is, actually)
        data = dict((name, self.data[name])
                    for name in self.variables)

        numpy.savez(NumpyDataset._openSingleFile(base), **data)

    @inheritdoc(NewROOT, gap='')
    def to_tree(self, treename, **options):
        """

        Parameters
        ----------
        treename: str
            Name of ROOT TTree to be created.
        options: none

        Returns
        -------
        ROOTDataset holding new ROOT TTree.
        """

        tree = root_numpy.array2tree(self.data, treename)
        from .rootdataset import ROOTDataset
        return ROOTDataset(tree, self.provenance+Formatting('ROOTDataset', treename))

    def __getitem__(self, object):

        if isinstance(object, str) and object in self.__dict__.keys():
            array = self.data[object].view(SkhepNumpyArray)
            array.name = object
            array.provenance = ObjectOrigin(repr(self))
            return array.copy()
            # return a copy in order to avoid unrecorded operation due to ndarray mutability
        elif isinstance(object, SkhepNumpyArray) and object.dtype == bool:
            data = self.data[ object ]
            provenance = self.provenance + Transformation("Subsetting dataset: {0}".format(object.name), object.name)
            return NumpyDataset(data, provenance)
        else:
            return self.data[object]

    def __setattr__(self, name, value):
        listofattributes = self.__dict__.keys()

        if isinstance(value, numpy.ndarray) and name != "_data" and name not in listofattributes :
            if value.shape != self.data .shape:
                raise ValueError('Arrays should have the same dimensions')
            else:
                from numpy.lib import recfunctions
                detail = getattr(value, 'provenance', None)
                data = recfunctions.append_fields(self.data , name, value, usemask=False)
                self._data  = data
                self._provenance += Transformation("Array {0} has been created".format(name), detail)
                self.__add_var(name)
        else:
            dict.__setattr__(self, name, value)

    def __setitem__(self, name, array):
        listofattributes = self.__dict__.keys()

        if isinstance(array, numpy.ndarray):
            if name not in listofattributes:
                self.__setattr__(name, array)
            elif name != "_data":
                if array.shape != self.data[name].shape:
                    raise ValueError('Arrays should have the same dimensions')
                if isinstance(array, SkhepNumpyArray) and array.provenance[0].detail == repr(self) and array.name == name:
                    self._data[name] = array.view(numpy.ndarray)
                    self._provenance += MultiProvenance(*array.provenance[1:])
                else:
                    array_name = getattr(array, 'name', None)
                    if array_name is None: array_name = repr(array)
                    detail = getattr(array, 'provenance', None)
                    self._data[name] = array
                    self._provenance += Transformation("Array {0} has been replaced by {1}".format(name, array_name), detail)
        else:
            raise ValueError('Not an array!')

    def __add_var(self, var):
        """Add columns of the array as property attributes."""
        def make_get_set(var):
            def getter(self):
                return self.__getitem__(var)
            def setter(self, array):
                self.__setitem__(var, array)
            return getter, setter

        setattr(NumpyDataset, var, property(*make_get_set(var)))
        self.__dict__[var] = self.data[var]

    def __repr__(self):
        """Class representation."""
        return "NumpyDataset{0}".format(repr(self.data).replace("array",""))

    def __str__(self):
        """Simple class representation."""
        return str(self.data)

    def __len__(self):
        return len(self.data)


def can_override_ufunc ( ):
    # check whether or not __array_ufunc__ can be overidden
    # possible for numpy version > 1.13

    version = numpy.__version__
    version = version.split(".")
    version = [int(v) for v in version]
    return version[0] >= 1. and version[1] >= 13.

def concatenate(NumpyDatasets):

    if isinstance(NumpyDatasets, (tuple, list)):
        if all(isinstance(nd, NumpyDataset) for nd in NumpyDatasets):

          data_concatenated = numpy.concatenate([nd.data for nd in NumpyDatasets])
          provenance = MultiProvenance(*[nd.provenance for nd in NumpyDatasets])
          provenance = MultiProvenance(ObjectOrigin(provenance.detail))

          return NumpyDataset( data_concatenated, provenance )

        else:
            raise TypeError("Can only concatenate a list/tuple of NumpyDatasets")
    else:
        raise TypeError("Can only concatenate a list/tuple of NumpyDatasets")


# -----------------------------------------------------------------------------
# SkhepNumpyArray
# -----------------------------------------------------------------------------

class SkhepNumpyArray(numpy.ndarray):

    def __new__(cls, inputarray, name=None, provenance=None):
        """Default constructor for SkhepNumpyArray.

        Parameters
        ----------
        input_array: numpy array, list or tuple.
        name: name of the variable.
        provenance: history of the variable.
        """

        instance = numpy.asarray(inputarray).view(cls)
        instance._name  = name

        if isinstance(provenance, MultiProvenance):
            instance._provenance = provenance.copy()
        else:
            if provenance is None:
                provenance = ObjectOrigin(repr(inputarray))
            if not isinstance(provenance, ( list, tuple ) ):
                provenance = [provenance]
            instance._provenance = MultiProvenance(*provenance)

        return instance

    @inheritdoc(numpy.ndarray)
    def __array_finalize__(self, obj):
        if obj is None: return
        self._name = getattr(obj, 'name', None)
        self._provenance = getattr(obj, 'provenance', MultiProvenance(ObjectOrigin("")))

    def copy(self):
        """Get a copy of the SkhepNumpyArray."""
        return SkhepNumpyArray(numpy.copy(self), name = self._name, provenance = self._provenance)

    @property
    def name(self):
        """Return the name of the variable inside the SkhepNumpyArray."""
        return self._name

    @property
    def provenance(self):
        """Return the provenance of the SkhepNumpyArray."""
        return self._provenance

    @provenance.setter
    def provenance(self, provenance):
        """Sets the provenance of the SkhepNumpyArray."""
        if isinstance(provenance, MultiProvenance):
            self._provenance = provenance
        elif isinstance(provenance, Provenance):
            self._provenance = MultiProvenance(provenance)
        else:
            raise ValueError("Inputs must be Provenance types!")

    @name.setter
    def name(self, name):
        """Sets the name of the variable inside the SkhepNumpyArray."""
        self._name = name

    @inheritdoc(numpy.ndarray)
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        args = []
        names_inputs = []
        for _input in inputs:
            if hasattr(_input, "name"):
                names_inputs.append(_input.name)
            else:
                names_inputs.append(repr(_input))
            array_input =  numpy.asarray(_input)
            if ufunc == numpy.divide:
                array_input = array_input.astype(float)
            args.append(array_input)

        outputs = kwargs.pop('out', None)
        if outputs is not None:
            out_args = []
            for o in outputs:
                array_out = numpy.asarray(o)
                if ufunc == numpy.divide:
                    array_out = array_out.astype(float)
                out_args.append(array_out)
            kwargs['out'] = tuple(out_args)

        name       = self.name
        provenance = self.provenance.copy()
        result     = getattr(ufunc, method)(*args, **kwargs)

        if method != "at":
          result = result.view(SkhepNumpyArray)
          result = self.__format_ufunc_result( provenance, result, ufunc, names_inputs, outputs)
          return result

    def __format_ufunc_result(self, provenance, result, ufunc, names_inputs, outputs = None):

        #comparisaton_operators
        comparisaton_operators = {"<ufunc 'greater'>": ">", "<ufunc 'less'>": "<",
                                  "<ufunc 'greater_equal'>": ">=", "<ufunc 'less_equal'>": "<=",
                                  "<ufunc 'equal'>": "==", "<ufunc 'not_equal'>": "!=",
                                  "<ufunc 'bitwise_and'>": "&", "<ufunc 'bitwise_or'>": "|"}

        #arithmetic operators
        arithmetic_operators =   {"<ufunc 'add'>": "+", "<ufunc 'subtract'>": "-",
                                  "<ufunc 'multiply'>": "*", "<ufunc 'divide'>": "/",
                                  "<ufunc 'true_divide'>": "/", "<ufunc 'power'>": "^"}

        functions = {"<ufunc 'mininum'>": "min", "<ufunc 'maximum'>": "max", "<ufunc 'sin'>": "sin",
                     "<ufunc 'sin'>": "cos", "<ufunc 'tan'>": "tan", "<ufunc 'arcsin'>": "arcsin",
                     "<ufunc 'arccos'>": "arccos", "<ufunc 'arctan'>": "arctan",
                     "<ufunc 'arctan2'>": "arctan2", "<ufunc 'sinh'>": "sinh", "<ufunc 'cosh'>": "cosh",
                     "<ufunc 'tanh'>": "tanh", "<ufunc 'arcsinh'>": "arcsinh",
                     "<ufunc 'arccosh'>": "arcosh", "<ufunc 'log'>": "log", "<ufunc 'log10'>": "log10",
                     "<ufunc 'log1p'>": "log1p", "<ufunc 'exp'>": "exp", "<ufunc 'expm1'>": "expm1",
                     "<ufunc 'absolute'>": "abs"}

        if ufunc != numpy.logical_and and ufunc != numpy.logical_or:

            #comparisaton_operators
            if str(ufunc) in comparisaton_operators.keys():
              name, provenance = self.__comparison_ufunc( comparisaton_operators[str(ufunc)], provenance, names_inputs)

            #functions
            elif str(ufunc) in functions.keys():
              name, provenance = self.__function_ufunc( functions[str(ufunc)], provenance, names_inputs)

            #arithemtic_operators
            elif ufunc == numpy.square:
              name, provenance = self.__square_ufunc( provenance, outputs)

            elif ufunc == numpy.sqrt:
              name, provenance = self.__sqrt_ufunc( provenance, outputs)

            elif str(ufunc) in arithmetic_operators.keys():
              name, provenance = self.__arithmetic_ufunc( arithmetic_operators[str(ufunc)], provenance, names_inputs, outputs)

            else:

                name, provenance = self.name, self.provenance

            if isinstance(result, SkhepNumpyArray):
                result.name       = name
                result.provenance = provenance

        return result

    def __function_ufunc( self, function, provenance, names_inputs):

      name = function + "("

      for ni in names_inputs:
        if ni == names_inputs[-1]:
          name += " {0} )".format(ni)
        else:
          name += " {0},".format(ni)

      provenance += Transformation(name)

      return name, provenance

    def __comparison_ufunc( self, operator, provenance, names_inputs ):
        lhs = names_inputs[0]
        rhs = names_inputs[1]

        if operator == "|":
            lhs = "("+lhs+")"
            rhs = "("+rhs+")"

        name = "{0} {1} {2}".format(lhs, operator, rhs)
        provenance = ObjectOrigin(name)
        return name, provenance

    def __format_operand( self, operand, operator):
        if operator == "*"   and any( _op in operand for _op in ["+","-","/"]):
            operand = "({0})".format( operand )
        elif operator == "/" and any( _op in operand for _op in ["+","-","*"]):
            operand = "({0})".format( operand )
        elif operator == "^" and any( _op in operand for _op in ["+","-","*","/"]):
            operand = "({0})".format( operand )

        return operand

    def __square_ufunc( self, provenance, outputs = None):

        name_input = self.name

        provenance += Transformation("{0} has been squared".format(name_input))

        if outputs is None:
            name_input = self.__format_operand( name_input, "^")
            name = "{0}**2".format(name_input)
        else:
            name = self.name

        return name, provenance

    def __sqrt_ufunc( self, provenance, outputs = None):

        name_input = self.name

        provenance += Transformation("{0} has been raised to the power of 0.5".format(name_input))

        if outputs is None:
          name_input = self.__format_operand( name_input, "^")
          name = "{0}**0.5".format(name_input)
        else:
          name = self.name

        return name, provenance

    def __arithmetic_ufunc( self, operator, provenance, names_inputs, outputs = None):
        lhs = names_inputs[0]
        rhs = names_inputs[1]

        # __add__
        if operator == "+":
            provenance += Transformation("{1} has been added to {0}".format(lhs, rhs))
        # __sub__
        elif operator == "-":
            provenance += Transformation("{1} has been subtracted to {0}".format(lhs, rhs))
        # __mul__
        elif operator == "*":
            provenance += Transformation("{0} has been multiplied by {1}".format(lhs, rhs))
        # __mul__
        elif operator == "/":
            provenance += Transformation("{0} has been divided by {1}".format(lhs, rhs))
        # power
        elif operator == "^":
            provenance += Transformation("{0} has been raised to the power of {1}".format(lhs, rhs))

        if outputs is None:
            lhs = self.__format_operand(lhs, operator)
            rhs = self.__format_operand(rhs, operator)
            if operator == "^":
                name = "{0}{1}{2}".format(lhs, operator, rhs)
            else:
                name = "{0} {1} {2}".format(lhs, operator, rhs)
        else:
            name = self.name

        return name, provenance

    ### overloading operators in case numpy version < 1.13, will

    if not can_override_ufunc():

        __array_priority__ = 2.0

        @inheritdoc(numpy.ndarray)
        def __array_wrap__(self, array, context=None):

            if context:
                _context = list(context)
                ufunc  = _context[0]
                inputs = _context[1]

                names_inputs = []
                for i in inputs:
                    names_inputs.append( getattr(i, 'name', repr(i)) )

                provenance = self.provenance.copy()
                array.view(SkhepNumpyArray)

                array = self.__format_ufunc_result( provenance, array, ufunc, names_inputs)

            return array

        def __iadd__(self, other):
            return self.__array_ufunc__(numpy.add, "__call__", self, other, out=(self,))

        def __isub__(self, other):
            return self.__array_ufunc__(numpy.subtract, "__call__", self, other, out=(self,))

        def __imul__(self, other):
            return self.__array_ufunc__(numpy.multiply, "__call__", self, other, out=(self,))

        def __idiv__(self, other):
            return self.__array_ufunc__(numpy.divide, "__call__", self, other, out=(self,))

        def __itruediv__(self, other):
            return self.__array_ufunc__(numpy.true_divide, "__call__", self, other, out=(self,))

        def __pow__(self, other):
            if isinstance(other, (int, float)) and other == 2.0:
                return self.__array_ufunc__(numpy.square, "__call__", self)
            else:
                return self.__array_ufunc__(numpy.power, "__call__", self, other)

        def __ipow__(self, other):
            if isinstance(other, (int, float)) and other == 2.0:
                return self.__array_ufunc__(numpy.square, "__call__", self, out=(self,))
            else:
                ufunc = numpy.power
                return self.__array_ufunc__(numpy.power, "__call__", self, other, out=(self,))




# -----------------------------------------------------------------------------
# Add Numpy methods to NumpyDataset in bulk.
# SkhepNumpyArray
# -----------------------------------------------------------------------------

#def addNumpyMethod(method):
#    def fn(self, name, *args, **kwds):
#        return method.__call__(self.data, *args, **kwds)
#
#    fn.__name__ = method.__name__
#    fn.__doc__ = method.__doc__
#    if sys.version_info[0]==2:  # ugly but works! TODO: find nicer way
#        setattr(NumpyDataset, method.__name__, MethodType(fn, None, NumpyDataset))
#    else:
#        setattr(NumpyDataset, method.__name__, MethodType(fn, NumpyDataset))
#
#  try:
#     addNumpyMethod(numpy.ndarray.__add__)
#     addNumpyMethod(numpy.ndarray.__mul__)
#     addNumpyMethod(numpy.ndarray.sum)
#     addNumpyMethod(numpy.ndarray.mean)
#  except ImportError:
#    pass



# Licensed under a 3-clause BSD style license, see LICENSE.
from __future__ import absolute_import
import numpy as np
import numpy.ma as ma
import collections
import matplotlib


def fill_between_steps(ax, x, y1, y2=0, step_where='pre', **kwargs):
    ''' Fill between for a step plot histogram.
    Modified from original code produced by T. Caswell (https://github.com/tacaswell).

    Parameters
    ----------
    ax : Axes
       The axes to draw to

    x : array-like
        Array/vector of index values.

    y1 : array-like or float
        Array/vector of values to be filled under.
    y2 : array-Like or float, optional
        Array/vector or bottom values for filled area. Default is 0.

    step_where : {'pre', 'post', 'mid'}
        where the step happens, same meanings as for `step`

    **kwargs will be passed to the matplotlib fill_between() function.

    Returns
    -------
    ret : PolyCollection
       The added artist

    '''

    # Modification to account for histogram-like bin-edges
    if len(x) == len(y1)+1 and len(y1) != 1:
        kwargs['linewidth'] = 0
        if isinstance(y2, collections.Container):
            y2_trunc = y2[0:2]
        else:
            y2_trunc = y2
        fill_between_steps(ax, x[0:2], y1[0:2], y2=y2_trunc, step_where='post', **kwargs)
        kwargs['label'] = None
        return fill_between_steps(ax, x[1:], y1, y2=y2, step_where='pre', **kwargs)

    # Account for case in which there is exactly one bin
    elif len(y1) == 1:
        y1 = y1[0]
        if isinstance(y2, collections.Container):
            y2 = y2[0]
        else:
            y2 = y2

    if step_where not in ['pre', 'post', 'mid']:
        raise ValueError("where must be one of {{'pre', 'post', 'mid'}} "
                         "You passed in {wh}".format(wh=step_where))

    # make sure y values are up-converted to arrays
    if np.isscalar(y1):
        y1 = np.ones_like(x) * y1

    if np.isscalar(y2):
        y2 = np.ones_like(x) * y2

    # temporary array for up-converting the values to step corners
    # 3 x 2N - 1 array

    vertices = np.vstack((x, y1, y2))

    # this logic is lifted from lines.py
    # this should probably be centralized someplace
    if step_where == 'pre':
        steps = ma.zeros((3, 2 * len(x) - 1), np.float)
        steps[0, 0::2], steps[0, 1::2] = vertices[0, :], vertices[0, :-1]
        steps[1:, 0::2], steps[1:, 1:-1:2] = vertices[1:, :], vertices[1:, 1:]

    elif step_where == 'post':
        steps = ma.zeros((3, 2 * len(x) - 1), np.float)
        steps[0, ::2], steps[0, 1:-1:2] = vertices[0, :], vertices[0, 1:]
        steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :-1]

    elif step_where == 'mid':
        steps = ma.zeros((3, 2 * len(x)), np.float)
        steps[0, 1:-1:2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
        steps[0, 2::2] = 0.5 * (vertices[0, :-1] + vertices[0, 1:])
        steps[0, 0] = vertices[0, 0]
        steps[0, -1] = vertices[0, -1]
        steps[1:, 0::2], steps[1:, 1::2] = vertices[1:, :], vertices[1:, :]
    else:
        raise RuntimeError("should never hit end of if-elif block for validated input")

    # un-pack
    xx, yy1, yy2 = steps

    # now to the plotting part:
    return ax.fill_between(xx, yy1, y2=yy2, **kwargs)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.dataset.numpydataset module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest
from pytest import approx
import os

import numpy as np

from skhep.utils.provenance import ObjectOrigin  # these are not imported automatically
from skhep.dataset.numpydataset import *
from skhep.dataset.selection import Selection


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------

ar = np.array([(1,1),(2,2),(3,3)],dtype=[('x',int), ('y',int)])

def test_constructors():
    ds = NumpyDataset(ar)
    a = SkhepNumpyArray([1,2,3])

def test_properties():
    ds = NumpyDataset(ar, ObjectOrigin('simple_array'))
    assert type(ds.data) == np.ndarray
    assert ds.provenance[0].__repr__() == '<ObjectOrigin>'
    assert ds.provenance.__repr__() == '<ObjectOrigin>'
    assert ds.mutable == True
    assert ds.immutable == False
    assert ds.transient == True
    assert ds.persistent == False
    with pytest.raises(NotImplementedError):
        ds.datashape

def test_methods():
    ds1 = NumpyDataset(ar)
    ds1.__repr__()
    assert ds1.__str__() == "[(1, 1) (2, 2) (3, 3)]"
    
    assert ds1.nentries == 3
    assert ds1.nevents  == 3
    
    assert ds1.variables == ["x", "y"]
    assert ds1.keys()    == ["x", "y"]

    ds1.to_file('npds.npz')
    ds2 = NumpyDataset.from_file('npds.npz')
    assert ds2.provenance[0].__repr__() == '<FileOrigin (1 file)>'
    os.remove('npds.npz')
    with pytest.raises(IOError):
        ds = NumpyDataset.from_file('non_existent_file')
    ds3 = ds1.copy()
    
    assert ds1['x'].tolist() == [1,2,3]
    assert ds1.x.tolist() == [1,2,3]
    assert ds1.x.name == "x"
    assert ds1.x.provenance[0].detail == repr(ds1)
    
    ds1.z = np.ones((3,))
    assert ds1['z'].tolist() == [1,1,1]
    assert ds1.z.tolist() == [1,1,1]
    assert ds1.provenance[-1].__repr__() == "<Transformation(Array z has been created)>"
    
    ds1['w'] = np.zeros((3,))
    assert ds1.provenance[-1].__repr__() == "<Transformation(Array w has been created)>"
    assert ds1['w'].tolist() == [0,0,0]
    assert ds1.w.tolist() == [0,0,0]
    
    with pytest.raises(ValueError):
        ds1.__setitem__('h',np.ones((4,)))
    with pytest.raises(ValueError):
        ds1.__setitem__('k',"array")
        
    ds1.z = ds1.w
    assert ds1.provenance[-1].__repr__() =="<Transformation(Array z has been replaced by w)>"
    ds1.w = np.array([6,7,8])
    assert ds1.provenance[-1].__repr__() =="<Transformation(Array w has been replaced by array([6, 7, 8]))>"
    
    
def test_transformations():

    ds1 = NumpyDataset(ar)
    ds1.x += 1
    assert repr(ds1.provenance[-1]) == "<Transformation(1 has been added to x)>"
    ds1.x *= 2
    assert repr(ds1.provenance[-1]) == "<Transformation(x has been multiplied by 2)>"
    ds1.x -= 3
    assert repr(ds1.provenance[-1]) == "<Transformation(3 has been subtracted to x)>"
    ds1.x /= 4
    assert repr(ds1.provenance[-1]) == "<Transformation(x has been divided by 4)>"
    ds1.r = (ds1.x**2 + ds1.y**2)**0.5
    ds1.expx = np.exp(ds1.x)
    
def test_operations():
    
    a = SkhepNumpyArray([1,2,3,4], name = "a")
    b = SkhepNumpyArray([0,1,2,3], name = "b")
    np.add.at(a, b, 1)
    assert a.all() == SkhepNumpyArray([2, 3, 4, 5]).all()
    
    c = np.exp(a)
    print(c[0])
    assert c[0] == approx(np.exp(2))
    assert c.provenance[-1].detail == "exp( a )"
    
def test_concatenate():
    
    ar1 = np.array([(1,1),(2,2),(3,3)],dtype=[('x',int), ('y',int)])
    ar2 = np.array([(4,-2),(0,7),(4,3)],dtype=[('x',int), ('y',int)])
    ar3 = np.array([(3,-1),(2,5),(1,1)],dtype=[('x',int), ('y',int)])
    
    ds1 = NumpyDataset(ar1)
    ds2 = NumpyDataset(ar2)
    ds3 = NumpyDataset(ar3)
    
    with pytest.raises(TypeError):
        ds = concatenate(ds1)
    with pytest.raises(TypeError):
        ds = concatenate([ds1,"a"])
    
    ds = concatenate([ds1, ds2, ds3])
    
    assert ds.nentries == ds1.nentries + ds2.nentries + ds3.nentries
    assert ds[-1][0] == 1 
    assert ds[-1][1] == 1
    
    assert repr(ds.provenance) == "<ObjectOrigin>"
#    details  = "0: {0}\n".format(ds1.provenance.detail)
#    details += "1: {0}\n".format(ds2.provenance.detail)
#    details += "2: {0}".format(ds3.provenance.detail)
#    assert ds.provenance.detail == details

    ds1_ = ds1.select("x > 1")
    ds2_ = ds2.select("x > 1")
    ds3_ = ds3.select("x > 1")
    ds_ = concatenate([ds1_, ds2_, ds3_])
    
    assert ds_.nentries == ds.select("x > 1").nentries
    
    assert repr(ds_.provenance) == "<ObjectOrigin>"
#    details  = "0: 0: {0}\n".format(ds1_.provenance[0].detail)
#    details += "   1: {0}\n".format(ds1_.provenance[1].detail)
#    details += "1: 0: {0}\n".format(ds2_.provenance[0].detail)
#    details += "   1: {0}\n".format(ds2_.provenance[1].detail)
#    details += "2: 0: {0}\n".format(ds3_.provenance[0].detail)
#    details += "   1: {0}".format(ds3_.provenance[1].detail)
#    assert ds_.provenance.detail == details

def test_selections():
    
    ar = np.array([(1,1),(2,2),(3,3)],dtype=[('x',int), ('y',int)])
    
    ds1 = NumpyDataset(ar)
    ds1.z = (ds1.x**2 + ds1.y**2)**0.5
    
    
    sel = Selection("x > 1")
    ds2 = ds1.select(sel)
    ds2.provenance[-1] == "<Transformation(Selection, (x > 1), applied)>"
    assert ds2.__str__() == ds1.select("x > 1").__str__()
    
    assert ds2.provenance[-1].detail == ("Selection, x > 1, applied (#events before = 3, "
                                         + "#events after = 2, Efficiency = 66.6667 +/- 27.2166 %)")
    
    ds3 = ds1[ds1.x > 1]
    a = ds1.x > 1
    print(ds3.provenance[-1])
    assert repr(ds3.provenance[-1]) == "<Transformation(Subsetting dataset: x > 1)>"
    ds4 = ds1.select(ds1.x > 1)
    assert repr(ds4.provenance[-1]) == "<Transformation(Subsetting dataset: x > 1)>"
    ds5 = ds1.select("(x > 1) & (y > 1)")
    ds6 = ds1.select("min(x,y) > 1")
    ds7 = ds1.select("max(x,y) > 1")
    ds8 = ds1.select("( x * y ) > 1")
    ds9 = ds1.select("max(x,y,z) > 1")
    with pytest.raises(ValueError):
        ds10 = ds1.copy()
        ds10.select(ds8.x)
#    dst = ds1.to_tree("DecayTree")
#    assert repr(dst.provenance[-1]) == "<Formatting to ROOTDataset(DecayTree)>


#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.math.geometry module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from skhep.math.geometry import *
from skhep.utils.py23 import *

import pytest
from pytest import approx

from math import pi

# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------
def test_geometry_constructors():
    
    v1  = Vector3D ( )
    p1  = Point3D  ( )
    assert str(p1)  == "Point3D(x=0.0,y=0.0,z=0.0)"
    assert repr(p1) == "<Point3D(x=0.0,y=0.0,z=0.0)>"
    
    l1  = Line3D   ( p1 , Vector3D ( 0 , 0 , 1 ) )
    assert str(l1)  == "Line3D({0},{1})"   .format(p1, Vector3D ( 0 , 0 , 1 ))
    assert repr(l1) == "<Line3D({0},{1})>" .format(p1, Vector3D ( 0 , 0 , 1 ))
    
    pl1 = Plane3D  ( p1 , Vector3D ( 0 , 0 , 1 ) )
    assert str(pl1)  == "Plane3D({0},{1})"   .format(p1, Vector3D ( 0 , 0 , 1 ))
    assert repr(pl1) == "<Plane3D({0},{1})>" .format(p1, Vector3D ( 0 , 0 , 1 ))

    l2  = Line3D  .from_points         ( p1 , Point3D(1,2,3) )
    pl2 = Plane3D .from_points         ( p1 , Point3D(1,2,3) , Point3D (3,2,1) )
    p2  = Point3D .frompoint           ( 1, 1, 1 )
    pl3 = Plane3D .from_line_and_point ( l1 , p2 )
    
    
    p3 = Point3D.fromiterable([1., 1., 1.])
    assert p3 == Point3D(1., 1., 1.)
    p4 = Point3D.fromcylindricalcoords(1., 0., 1.)
    assert p4 == Point3D(1., 0., 1.)
    assert p4.rho   == 1.
    p5 = Point3D.fromsphericalcoords(1.0, 0., 0.)
    assert p5 == Point3D(0., 0., 1.)
    assert p5.theta == 0.
    assert p5.phi   == 0.
    
    v2  = Vector3D ( 1, 1, 1)
    with pytest.raises(NotImplementedError):
        Line3D.__init__(Line3D(), p1, p1)
    with pytest.raises(NotImplementedError):
        Line3D.__init__(Line3D(), v2, v2)
    with pytest.raises(ValueError):
        Line3D.__init__(Line3D(), p1, v1)
        
    with pytest.raises(NotImplementedError):
        Plane3D.__init__(Plane3D(), p1, p1)
    with pytest.raises(NotImplementedError):
        Plane3D.__init__(Plane3D(), v2, v2)
    with pytest.raises(ValueError):
        Plane3D.__init__(Plane3D(), p1, v1)

def test_operators ():
    v1 = Vector3D(1,2,3)
    v2 = Vector3D(3,2,1)
    p1 = Point3D (0,0,1)
    p2 = Point3D (0,1,0)
    line1 = Line3D( p1 , v1)
    line2 = Line3D( p2 , v2)
    plane1 = Plane3D( p1 , v1)
    plane2 = Plane3D( p2 , v2)
    
    assert p1 + v1 == Point3D (1,2,4)
    assert v1 + p1 == Point3D (1,2,4)
    
    assert p1 - v1 == Point3D (-1,-2,-2)
    assert p1 - p2 == Vector3D(0,-1,1)
    
    p1 -= v1
    p1 += v1
    
    assert p1 == p1
    assert p1 != p2
    
    assert line1 == line1
    assert line1 != line2
    
    assert plane1 == plane1
    assert plane1 != line2


def test_contains():
    line  = Line3D   ( Point3D() , Vector3D(0,0,1) )
    plane = Plane3D  ( Point3D() , Vector3D(0,0,1) )

    assert Point3D(0,0,-10) in line
    assert Point3D(1,0,-10) not in line

    assert Point3D(1,0,0) in plane
    assert Point3D(0,1,0) in plane
    assert Point3D(0,0,2) not in plane

    linex = Line3D(  Point3D() , Vector3D(1,0,0) )
    liney = Line3D(  Point3D() , Vector3D(0,1,0) )
    linez = Line3D(  Point3D() , Vector3D(0,0,1) )

    assert linex in plane
    assert liney in plane
    assert linez not in plane


def test_distance():
    
    with pytest.raises(NotImplementedError):
        Point3D.distance(Point3D(), 2.)
    with pytest.raises(NotImplementedError):
        Line3D.distance(Line3D(), 2.)
    with pytest.raises(NotImplementedError):
        Plane3D.distance(Plane3D(), 2.)
    
    p0 = Point3D()
    p1 = Point3D(1,0,0)

    assert p0.distance(p0) == 0
    assert p0.distance(p1) == 1

    line = Line3D( Point3D() , Vector3D(0,0,1) )

    assert p0.distance(line) == 0
    assert p1.distance(line) == 1

    assert line.distance(p0) == 0
    assert line.distance(p1) == 1

    plane = Plane3D  ( Point3D(1,1,0) ,  Vector3D(0,0,1) )

    assert plane.distance(plane) == 0
    assert plane.distance(p0) == 0

    line1 = Line3D ( Point3D(1,2,3) , Vector3D (4,5,6) )
    line2 = Line3D ( Point3D(0,0,1) , Vector3D (1,1,0) )

    assert plane.distance(line1) == 0
    assert plane.distance(line2) == 1
    assert line1.distance(plane) == 0
    assert line2.distance(plane) == 1
    
    line3 = Line3D ( Point3D(0,0,2) , Vector3D (1,1,0) )
    
    assert line3.distance(line2) == 1
    
def test_angle():
    
    p0 = Point3D()
    
    with pytest.raises(NotImplementedError):
        Line3D.angle(Line3D(), p0)
    with pytest.raises(NotImplementedError):
        Plane3D.angle(Plane3D(), p0)
    
    line0  = Line3D   ( Point3D() , Vector3D(0,0,1) )
    line1  = Line3D   ( Point3D(0,0,1) , Vector3D(0,1,0) )

    plane0 = Plane3D  ( Point3D() , Vector3D(0,0,1) )
    plane1 = Plane3D  ( Point3D(0,0,1) , Vector3D(0,1,0) )
    
    assert line0.angle ( line1 ) == pi/2.
    assert line1.angle ( line0 ) == pi/2.
    assert line1.angle ( line1 ) == 0.
    assert line0.angle ( plane0 ) == 0.
    assert line0.angle ( plane1 ) == pi/2.
    assert line1.angle ( plane0 ) == pi/2.
    assert line1.angle ( plane1 ) == 0.
    
    assert plane0.angle ( plane1 ) == pi/2.
    assert plane1.angle ( plane0 ) == pi/2.
    assert plane1.angle ( plane1 ) == 0.
    
    assert plane0.angle ( line0 ) == 0.
    assert plane0.angle ( line1 ) == pi/2.
    assert plane1.angle ( line0 ) == pi/2.
    assert plane1.angle ( line1 ) == 0.

    
def test_intersect():

    with pytest.raises(NotImplementedError):
        Line3D.intersect(Line3D(), 1)
    with pytest.raises(NotImplementedError):
        Plane3D.intersect(Plane3D(), 1)
    
    p0 = Point3D()
    p1 = Point3D(0,0,1)
    
    line0  = Line3D   ( Point3D() , Vector3D(0,0,1) )
    line1  = Line3D   ( Point3D(0,0,1) , Vector3D(0,1,0) )

    plane0 = Plane3D  ( Point3D() , Vector3D(0,0,1) )
    plane1 = Plane3D  ( Point3D(0,0,1) , Vector3D(0,1,0) )
    
    assert line0.intersect( p0 ) == p0
    assert line0.intersect( p1 ) == p1
    assert line1.intersect( p0 ) == None
    assert line1.intersect( p1 ) == p1
    
    assert line0.intersect( line1 ) == Point3D(0,0,1)
    assert line1.intersect( line0 ) == Point3D(0,0,1)
    
    assert line0.intersect( plane0 ) == Point3D()
    assert line0.intersect( plane1 ) == line0
    assert line1.intersect( plane0 ) == None
    assert line1.intersect( plane1 ) == Point3D(0,0,1)
        
    assert plane0.intersect( p0 ) == p0
    assert plane0.intersect( p1 ) == None
    assert plane1.intersect( p0 ) == p0
    assert plane1.intersect( p1 ) == p1
    
    assert plane0.intersect( line0 ) == Point3D()
    assert plane0.intersect( line1 ) == None
    assert plane1.intersect( line0 ) == line0
    assert plane1.intersect( line1 ) == Point3D(0,0,1)
    
    assert plane0.intersect( plane1 ) == Line3D( Point3D() , Vector3D(1,0,0) )
    assert plane1.intersect( plane0 ) == Line3D( Point3D() , Vector3D(1,0,0) )
    
    assert plane0.intersect( plane0 ) == None
    line2  = Line3D   ( Point3D(0,1,0) , Vector3D(0,0,1) )
    assert line2.intersect( line0 ) == None


# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Module for algorithms and methods used to model distributions.

This module contains in particular:

  * Bayesian Blocks binning algorithm.
"""

#-----------------------------------------------------------------------------
# Import statements
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from .bayesian_blocks import bayesian_blocks



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.math.vectors module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest
from pytest import approx

from skhep.math.vectors import *
from skhep.utils.py23 import *
from skhep.utils.exceptions import *
from math import pi, sqrt


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------
def test_vectors_constructors():
    with pytest.raises(TypeError):
        Vector3D.fromiterable(1.)
    with pytest.raises(ValueError):
        Vector3D.fromiterable([1.])
    with pytest.raises(ValueError):
        Vector3D.fromiterable(['str1','str2','str3'])
    #
    v1 = Vector3D()
    assert str(v1) == str((0., 0., 0.))
    assert repr(v1) == "<Vector3D (x=0.0,y=0.0,z=0.0)>"
    assert str(v1) == str(Vector3D.origin())
    v2 = Vector3D(1.,1.,1.)
    assert str(v2) == str((1., 1., 1.))
    v3 = Vector3D.fromvector(v1)
    assert str(v3) == str((0., 0., 0.))
    v4 = Vector3D.fromiterable([1.0, 1.0, 1.0])
    assert str(v4) == str((1., 1., 1.))
    v5 = Vector3D.fromcylindricalcoords(1., 0., 1.)
    assert v5 == Vector3D(1., 0., 1.)
    assert v5.rho == 1.
    v6 = Vector3D.fromcylindricalcoords(0.5, pi/2, 0.)
    assert v6 == Vector3D(0., 0.5, 0.)
    v7 = Vector3D.fromsphericalcoords(1.0, 0., 0.)
    assert v7 == Vector3D(0., 0., 1.)
    assert v7.r == 1.
    assert v7.theta() == 0.
    v8 = Vector3D.fromsphericalcoords(1.0, 0., pi/2)
    assert v8 == Vector3D(0., 0., 1.)
    v9 = Vector3D.fromsphericalcoords(2.0, pi/2, pi/4)
    assert v9 == Vector3D(sqrt(2), sqrt(2), 0.)
    v10 = Vector3D.frompoint( 2., 2., 2.)
    assert str(v10) == str((2., 2., 2.))
    #
    with pytest.raises(TypeError):
        LorentzVector.fromiterable(1)
    with pytest.raises(ValueError):
        LorentzVector.fromiterable([1])
    with pytest.raises(ValueError):
        LorentzVector.fromiterable(['str1', 'str2', 'str3', 'str4'])
    #
    lv1 = LorentzVector()
    assert str(lv1) == str((0., 0., 0., 0.))
    assert repr(lv1) == "<LorentzVector (x=0.0,y=0.0,z=0.0,t=0.0)>"
    lv2 = LorentzVector(1., 1., 1., 1.)
    assert str(lv2) == str((1., 1., 1., 1.))
    lv3 = LorentzVector.from4vector(lv1)
    assert str(lv3) == str((0., 0., 0., 0.))
    lv4 = LorentzVector.from3vector(v2, 2.)
    assert str(lv4) == str((1., 1., 1.,  2.))
    lv5 = LorentzVector.fromiterable([1., 1., 1., 0.])
    assert str(lv5) == str((1., 1., 1., 0.))

def test_containers_properties():
    with pytest.raises(IndexError):
        Vector3D.__setitem__(Vector3D(), 3, 1.)
    with pytest.raises(IndexError):
        Vector3D.__getitem__(Vector3D(), 3)
    #
    v1 = Vector3D()
    v1.x = 5.; v1.y = -5.; v1.z = 10
    assert v1 == Vector3D(5., -5.,  10.)
    v1.set(-5., 5., 10.)
    assert v1 == Vector3D(-5., 5.,  10.)
    v1[0] = 1.
    assert v1 == Vector3D(1., 5.,  10.)
    v1[1] = 1.
    assert v1 == Vector3D(1., 1.,  10.)
    v1[2] = 1.
    assert v1 == Vector3D(1., 1.,  1.)
    assert v1[0] == 1.
    assert v1[1] == 1.
    assert v1[2] == 1.
    assert len(v1) == 3.
    assert v1.tolist() == [1., 1.,  1.]
    assert list(v1) == [1., 1.,  1.]
    assert [v for v in v1] == [1., 1.,  1.]
    #
    with pytest.raises(IndexError):
        LorentzVector.__setitem__(LorentzVector(), 4, 1.)
    with pytest.raises(IndexError):
        LorentzVector.__getitem__(LorentzVector(), 4)
    #
    lv1 = LorentzVector()
    lv1.x = 5.; lv1.y = -5.; lv1.z = 10; lv1.t = 2.
    assert lv1 == LorentzVector(5., -5., 10.,  2.)
    lv1.px = 5.; lv1.py = 5.; lv1.pz = -10; lv1.e = 2.
    assert lv1 == LorentzVector(5., 5., -10.,  2.)
    lv1.set(-5., 5., 10., -2.)
    assert lv1 == LorentzVector(-5., 5., 10.,  -2.)
    lv1[0] = 1.
    assert lv1 == LorentzVector(1., 5., 10.,  -2.)
    lv1[1] = 1.
    assert lv1 == LorentzVector(1., 1., 10.,  -2.)
    lv1[2] = 1.
    assert lv1 == LorentzVector(1., 1., 1.,  -2.)
    lv1[3] = 1.
    assert lv1 == LorentzVector(1., 1., 1.,  1.)
    assert lv1[0] == 1.
    assert lv1[1] == 1.
    assert lv1[2] == 1.
    assert lv1[3] == 1.
    assert len(lv1) == 4.
    assert lv1.tolist() == [1., 1., 1.,  1.]
    assert list(lv1) == [1., 1., 1.,  1.]
    assert [v for v in lv1] == [1., 1., 1.,  1.]

def test_vectors_operators():
    with pytest.raises(InvalidOperationError):
        Vector3D.__iadd__(Vector3D(), "a")
    with pytest.raises(InvalidOperationError):
        Vector3D.__isub__(Vector3D(), "a")
    with pytest.raises(InvalidOperationError):
        Vector3D.__imul__(Vector3D(), Vector3D())
    with pytest.raises(ZeroDivisionError):
        Vector3D.__div__(Vector3D(), 0.0)
    with pytest.raises(InvalidOperationError):
        Vector3D.__idiv__(Vector3D(), Vector3D())
    with pytest.raises(ZeroDivisionError):
        Vector3D.__idiv__(Vector3D(), 0.0)
    #
    v1, v2 = Vector3D(0., 0., 0.), Vector3D(1., 1., 1.)
    v3, v4 = Vector3D(2., 2., 2.), Vector3D(3., 3., 3.)
    v5 = Vector3D(1., 2., 3.)
    assert v1 == 0.
    assert v1 + v2 == Vector3D(1., 1., 1.)
    assert v1 - v2 == Vector3D(-1., -1., -1.)
    assert v1 - v2 == -1.*Vector3D(1., 1., 1.)
    assert v1 * 2. == Vector3D(0., 0., 0.)
    assert v2 * 2. == Vector3D(2., 2., 2.)
    assert 2. * v1 == Vector3D(0., 0.,  0.)
    assert 2. * v2 == Vector3D(2., 2., 2.)
    assert v1 * v2 == 0.
    assert v2 * v1 == 0.
    assert v3 * v4 == v3.dot(v4), "Check commutativity"
    assert v3 * v4 == v4.dot(v3), "Check commutativity"
    assert v4 * v3 == v3.dot(v4), "Check commutativity"
    assert v4 * v3 == v4.dot(v3), "Check commutativity"
    assert v3 * v4 == 18.
    assert v3 / 2. == v2
    assert v4 / 3. == v2
    v1 *= 2.
    v2 *= 2.
    assert v1 == Vector3D(0., 0., 0.)
    assert v2 == Vector3D(2., 2., 2.)
    v1 /= 2.
    v2 /= 2.
    assert v1 == Vector3D(0., 0., 0.)
    assert v2 == Vector3D(1., 1., 1.)
    assert (v1 + v2) * v3 == 6., "Check operations combination"
    assert (v2 - v1) * v3 == 6., "Check operations combination"
    assert v3 * (v1 - v2) == -6., "Check operations combination"
    assert 18. / (v3 * v4) == 1., "Check operations combination"
    assert  v4 / (v3 * v2) == Vector3D(0.5, 0.5, 0.5)
    assert v2.cross(v2) == Vector3D(0., 0., 0.)
    assert v2.cross(v5) == Vector3D(1., -2., 1.)
    assert v5.cross(v2) == -1 * v2.cross(v5)
    assert v2 == v2
    assert v2 != v1
    assert not v2 != v2
    assert not v2 == v1
    assert not v1 == "a"
    assert not v1 == 1
    #
    with pytest.raises(InvalidOperationError):
        LorentzVector.__iadd__(LorentzVector(), "a")
    with pytest.raises(InvalidOperationError):
        LorentzVector.__isub__(LorentzVector(), "a")
    with pytest.raises(InvalidOperationError):
        LorentzVector.__imul__(LorentzVector(), LorentzVector())
    with pytest.raises(ZeroDivisionError):
        LorentzVector.__div__(LorentzVector(), 0.0)
    with pytest.raises(InvalidOperationError):
        LorentzVector.__idiv__(LorentzVector(), LorentzVector())
    with pytest.raises(ZeroDivisionError):
        LorentzVector.__idiv__(LorentzVector(), 0.0)
    #
    lv1, lv2 = LorentzVector(0., 0., 0., 0.), LorentzVector(1., 1., 1., 0.)
    lv3, lv4 = LorentzVector(2., 2., 2., 1.), LorentzVector(3., 3., 3., 1.)
    lv5 = LorentzVector(1., 1., 1., 6.)
    assert lv1 == 0.
    assert lv1 + lv2 == LorentzVector(1., 1., 1.,  0.)
    assert lv1 - lv2 == LorentzVector(-1., -1., -1.,  0.)
    assert lv1 * 2. == LorentzVector(0., 0., 0.,  0.)
    assert lv2 * 2. == LorentzVector(2., 2., 2.,  0.)
    assert 2. * lv1 == LorentzVector(0., 0., 0.,  0.)
    assert 2. * lv2 == LorentzVector(2., 2., 2.,  0.)
    assert lv2 * lv1 == 0.
    assert lv1 * lv2 == 0.
    assert lv3 * lv4 == lv3.dot(lv4)
    assert lv3 * lv4 == lv4.dot(lv3)
    assert lv4 * lv3 == lv3.dot(lv4)
    assert lv4 * lv3 == lv4.dot(lv3)
    assert lv3 * lv4 == -17.
    assert lv3 / 2. == LorentzVector(1., 1., 1.,  0.5)
    assert lv4 / 3. == LorentzVector(1., 1., 1.,  1./3)
    assert lv3 * lv5 == 0.0
    lv1 *= 2.
    lv2 *= 2.
    assert lv1 == LorentzVector(0., 0., 0.,  0.)
    assert lv2 == LorentzVector(2., 2., 2.,  0.)
    lv1 /= 2.
    lv2 /= 2.
    assert lv1 == LorentzVector(0., 0., 0.,  0.)
    assert lv2 == LorentzVector(1., 1., 1.,  0.)
    assert lv2 == lv2
    assert lv2 != lv1
    assert not lv2 != lv2
    assert not lv2 == lv1
    assert not lv1 == "a"
    assert not lv1 == 1

def test_vectors_rotations():
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, 1)
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, [1,2])
    with pytest.raises(TypeError):
        Vector3D.rotate(Vector3D(), pi, 1, 2, 3, 4)
    with pytest.raises(ValueError):
        Vector3D.rotate(Vector3D(), pi, 0, 1, 'a')
    with pytest.raises(ValueError):
        Vector3D.rotate(Vector3D(), pi, ['a','b',3])
    #
    v1 = Vector3D.fromcylindricalcoords(1., 0., 0.)
    assert v1.phi() == 0.0
    assert v1.rotatez(pi/2).phi() == pi/2
    assert v1.rotatez(pi/2) == Vector3D(0., 1.,  0.)
    assert v1.rotatey(pi).phi() == pi
    assert v1.rotatey(-pi).phi() == pi
    assert v1.rotatey(pi) == Vector3D(-1., 0.,  0.)
    assert v1.rotatey(-pi) == Vector3D(-1., 0.,  0.)
    assert v1.rotatex(pi).phi() == 0.
    assert v1.rotatex(pi) == Vector3D(1., 0.,  0.)
    v2 = Vector3D.fromsphericalcoords(1.0, pi/2, pi/2)
    assert v2.phi() == pi/2
    assert v2.theta() == pi/2
    assert v2.rotatex(pi).phi() == -pi/2
    assert v2.rotatex(pi).theta() == pi/2
    assert v2.rotatex(pi)  == Vector3D(0., -1.,  0.)
    v3 = Vector3D.fromsphericalcoords(1.0, pi/4, pi/4)
    angle = v2.angle(v3)
    axis = v2.cross(v3)
    assert v2.rotate(angle,  axis) ==  v3
    assert v2.rotate(-angle,  -1.*axis) ==  v3
    #
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, 1)
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, [1,2])
    with pytest.raises(TypeError):
        LorentzVector.rotate(LorentzVector(), pi, 1, 2, 3, 4)
    with pytest.raises(ValueError):
        LorentzVector.rotate(LorentzVector(), pi, 0, 1, 'a')
    with pytest.raises(ValueError):
        LorentzVector.rotate(LorentzVector(), pi, ['a','b',3])
    #
    lv1 = LorentzVector.from3vector(v1, 1.)
    assert lv1.phi() == 0.
    assert lv1.rotatez(pi/2).phi() == pi/2
    assert lv1.rotatez(pi/2) == LorentzVector(0., 1., 0.,  1.)
    assert lv1.rotatey(pi).phi() == pi
    assert lv1.rotatey(-pi).phi() == pi
    assert lv1.rotatey(pi) == LorentzVector(-1., 0., 0.,  1.)
    assert lv1.rotatey(-pi) == LorentzVector(-1., 0., 0.,  1.)
    assert lv1.rotatex(pi).phi() == 0.
    assert lv1.rotatex(pi) == LorentzVector(1., 0., 0.,  1.)
    lv2 = LorentzVector.from3vector(v2, 2.0)
    assert lv2.phi() == pi/2
    assert lv2.theta() == pi/2
    assert lv2.rotatex(pi).phi() == -pi/2
    assert lv2.rotatex(pi).theta() == pi/2
    assert lv2.rotatex(pi)  == LorentzVector(0., -1., 0.,  2.0)
    lv3 = LorentzVector.from3vector(v3, 2.0)
    assert lv2.rotate(angle, axis) ==  lv3
    assert lv2.rotate(-angle, -1.*axis) == lv3

def test_3Dvectors_properties():
    v0 = Vector3D()
    v1, v2 = Vector3D(1., 1., 1.), Vector3D(2., 2., 2.)
    v3, v4 = Vector3D(-1., -1., -1.), Vector3D(-2., -2., -2.)
    v5, v6 = Vector3D(1., 1., 0.), Vector3D(0., 0., 2.)
    assert not v0
    assert v1
    assert v1.mag2 == 3.
    assert v1.mag == sqrt(3.)
    assert v2.mag2 == 12.
    assert v2.mag == sqrt(12.)
    assert v1.unit().mag == 1.
    v7 = v1.unit()
    assert v1.unit() == v7.unit()
    assert v1.isparallel(v2) == True
    assert v2.isparallel(v1) == True
    assert v1.isantiparallel(v2) == False
    assert v2.isantiparallel(v1) == False
    assert v1.isantiparallel(v3) == True
    assert v2.isantiparallel(v4) == True
    assert v1.isparallel(v3) == False
    assert v2.isparallel(v4) == False
    assert v1.isopposite(v3) == True
    assert v1.isopposite(v4) == False
    assert v2.isopposite(v4) == True
    assert v2.isopposite(v3) == False
    assert v1.isperpendicular(v2) == False
    assert v2.isperpendicular(v1) == False
    assert v5.isperpendicular(v6) == True
    assert v6.isperpendicular(v5) == True

def test_lorentz_vectors_properties():
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), 1)
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), [1,2])
    with pytest.raises(TypeError):
        LorentzVector.boost(LorentzVector(), 1, 2, 3, 4)
    with pytest.raises(ValueError):
        LorentzVector.boost(LorentzVector(), 0, 1, 'a')
    with pytest.raises(ValueError):
        LorentzVector.boost(LorentzVector(), ['a','b',3])
    #
    lv0 = LorentzVector()
    lv1, lv2 = LorentzVector(1., 1., 1., 1.), LorentzVector(1., 1., 1., 2.)
    assert lv1.boostvector == Vector3D(1., 1.,  1.)
    assert lv1.beta == sqrt(3.)
    assert lv2.boostvector == Vector3D(0.5, 0.5,  0.5)
    lv3 = LorentzVector(0., 0., 1., 0.)
    beta = 0.05
    gamma = 1/sqrt(1 - beta**2)
    lv4 = lv3.boost(0,0,beta)
    assert lv4.z == lv3.z * gamma,  "Check length contraction"
    assert lv4.x == lv3.x
    assert lv4.y == lv3.x
    assert lv4.t == gamma * (lv3.t - beta * lv3.z)
    assert lv4 == LorentzVector(lv3.x, lv3.y, lv3.z * gamma,  gamma * (lv3.t - beta * lv3.z))
    lv5 = LorentzVector(0., 0., 0., 1.)
    assert lv5.beta == 0.
    assert lv5.gamma == 1.
    assert lv5.boost(0,0,0) == lv5
    lv6 = lv5.boost([0,beta,0])
    assert lv6.x == lv5.x
    assert lv6.z == lv5.z
    assert lv6.t == lv5.t * gamma,  "Check time dilation"
    assert lv6.y == gamma * (lv5.y - beta*lv5.t)
    assert lv6 == LorentzVector(lv5.x, gamma * (lv5.y - beta*lv5.t), lv5.z,  lv5.t * gamma)
    assert lv1.isspacelike() == True
    assert lv1.istimelike() == False
    assert lv1.islightlike() == False
    assert lv2.isspacelike() == False
    assert lv2.istimelike() == True
    assert lv2.islightlike() == False
    lv7 = LorentzVector(1., 1., 1., sqrt(3.))
    assert lv7.isspacelike() == False
    assert lv7.istimelike() == False
    assert lv7.islightlike() == True
    #
    p1 = LorentzVector()
    p1.setpxpypzm(5.,5.,10.,5)
    assert p1.px == 5.
    assert p1.py == 5.
    assert p1.pz == 10.
    assert p1.m == 5.
    assert p1.mass == 5.
    assert p1.m2 == 25.
    assert p1.mass2 == 25.
    assert p1.pt == sqrt(p1.px**2 + p1.py**2)
    assert p1.p == sqrt(p1.px**2 + p1.py**2 + p1.pz**2)
    assert p1.p == sqrt(p1.pt**2 + p1.pz**2)
    assert p1.e == sqrt(p1.m**2 + p1.p**2)
    assert p1.beta == p1.p / p1.e
    p2 = p1.boost(p1.px/p1.e, p1.y/p1.e, p1.z/p1.e)
    assert p2  == p1.boost(p1.boostvector)
    assert p2.p == 0.
    assert p2.m == approx(5.)
    p3 = LorentzVector()
    p3.setpxpypze(5.,5.,10.,20)
    assert p3.px == 5.
    assert p3.py == 5.
    assert p3.pz == 10.
    assert p3.e == 20.
    assert p3.m == sqrt(p3.e**2 - p3.p**2)
    assert p3.beta == p3.p / p3.e
    p4 = LorentzVector()
    p4.setptetaphim(10.,3.5,pi/3,5.)
    assert p4.pt == approx(10.)
    assert p4.eta == approx(3.5)
    assert p4.phi() == approx(pi/3)
    assert p4.m == approx(5.)
    p5 = LorentzVector()
    p5.setptetaphie(10.,3.9,-2*(pi/3),20.)
    assert p5.pt == approx(10.)
    assert p5.eta == approx(3.9)
    assert p5.pseudorapidity == approx(3.9)
    assert p5.phi() == approx(-2*(pi/3))
    assert p5.e == approx(20.)
    assert p5.deltaeta(p4) == approx(0.4)
    assert p4.deltaeta(p5) == approx(-0.4)
    assert p5.deltaphi(p4) == approx(-pi)
    assert p4.deltaphi(p5) == approx(pi)
    assert p5.deltar(p4) == approx(sqrt(0.4**2 + pi**2))
    assert p4.deltar(p5) == approx(sqrt(0.4**2 + pi**2))
    p6 = LorentzVector()
    p6.setptetaphie(10.,3.9,-pi,20.)
    assert p6.deltaphi(p4) == approx(2*(pi/3))
    assert p4.deltaphi(p6) == approx(-2*(pi/3))
    p7 = LorentzVector()
    p7.setpxpypzm(5.,5.,5.,0.)
    assert p7.beta == 1.
    assert p7.gamma == 10E10,  "Gamma of the photons is +inf"
    assert p7.p == p7.e,  "Momentum = Energy for photons"
    assert p7.pseudorapidity == p7.rapidity
    assert p7.pt == p7.et,  "Transverse Momentum = Transverse Energy for photons"
    assert p7.pt == approx(p7.mt), "Transverse Momentum = Transverse Mass for photons"
    assert p7.perp2 == approx(p7.mt2)
    assert p7.islightlike() == True
    p8 = p1 + p7
    p9, p10 = p1.boost(p8.boostvector), p7.boost(p8.boostvector)
    assert p9.p == approx(p10.p), "Check boost to the C.O.M frame"
    p3_9, p3_10 = p9.vector, p10.vector
    assert p3_9.isopposite(p3_10)
    p8 = LorentzVector()
    p8.setptetaphie(10.,2E2,-2*(pi/3),20.) #sinh(eta) diverge quickly
    assert p8.theta() == 0.0
    assert p8.eta == 10E10



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Submodule for decorators
========================

.. note:: not meant for user code in general, though possible.
"""


def inheritdoc(cls, gap="\n"):
    """
    Decorator to automatize the inheritance of documentation from a class method.

    Example
    -------

    >>> from skhep.utils.decorators import inheritdoc
    >>> class ADerivedClass(ABaseClass):                 # doctest: +SKIP
    ...    @inheritdoc(ABaseClass)                       # doctest: +SKIP
    ...    def amethod(self): pass                       # doctest: +SKIP
    """
    def _fn(fn):
        if fn.__name__ in cls.__dict__:
            if fn.__doc__ is None:
                fn.__doc__ = cls.__dict__[fn.__name__].__doc__
            else:
                fn.__doc__ = cls.__dict__[
                    fn.__name__].__doc__.strip() + gap + fn.__doc__.strip()
        return fn
    return _fn



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.dataset.rootdataset module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest

import os
from random import gauss

# Obviously, skip all tests if the ROOT library is not available!
ROOT = pytest.importorskip('ROOT')

from ROOT import gROOT, TTree, TChain, TFile, AddressOf

from skhep.dataset.rootdataset import *
from skhep.dataset.selection import Selection

# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------

gROOT.ProcessLine(
"struct pyTTree {\
Int_t run;\
Int_t evt;\
Float_t x;\
Float_t y;\
Float_t z;\
}"
)
from ROOT import pyTTree
pytree = pyTTree()

def create_simple_tree(nevents):
    """Create a simple TTree with a couple of branches."""
    t = TTree('aTTree','A TTree')
    t.Branch('run', AddressOf(pytree, 'run'),'run/I')
    t.Branch('evt', AddressOf(pytree, 'evt'),'evt/I')
    t.Branch('x', AddressOf(pytree, 'x'),'x/F')
    t.Branch('y', AddressOf(pytree, 'y'),'y/F')
    t.Branch('z', AddressOf(pytree, 'z'),'z/F')
    for i in range(nevents):
        pytree.run = 1 if (i<500) else 2
        pytree.evt = i
        pytree.x = gauss(0., 10.)
        pytree.y = gauss(0, 10.)
        pytree.z = gauss(5., 50.)
        t.Fill()
    return t

def create_root_files():
    tree1 = create_simple_tree(200)
    tree2 = create_simple_tree(300)
    tname = tree1.GetName()
    f1 = TFile('tree1.root','create')
    tree1.Write()
    f1.Close()
    f2 = TFile('tree2.root','create')
    tree2.Write()
    f2.Close()
    fnames = ['tree1.root', 'tree2.root']
    del f1, f2
    del tree1, tree2
    return fnames, tname

tree = create_simple_tree(1000)
filenames, treename = create_root_files()
chain = TChain(treename)
for fn in filenames:
    chain.Add(fn)

def test_constructors():
    ds1 = ROOTDataset(tree)
    ds2 = ROOTDataset(chain)
    del ds1, ds2

def test_properties():
    ds1 = ROOTDataset(tree)
    assert ds1.provenance.__repr__() == '<ObjectOrigin>'
    assert ds1.mutable == True
    assert ds1.immutable == False
    assert ds1.transient == False
    assert ds1.persistent == True
    ds2 = ROOTDataset(chain)
    assert ds2.provenance.__repr__() == '<FileOrigin (2 files)>'
    with pytest.raises(NotImplementedError):
        ds1.datashape
    del ds1, ds2

def test_methods():
    ds1 = ROOTDataset(tree)
    ds1.to_file('new_tree.root')
    assert os.path.isfile('new_tree.root') == True
    del ds1
    ds2 = ROOTDataset(chain)
    ds2.to_file('new_chain.root')
    assert os.path.isfile('new_chain.root') == True
    del ds2
    ds3 = ROOTDataset.from_file('new_tree.root', treename='aTTree')
    del ds3
    ds4 = ROOTDataset.from_file('new_tree.root')
    ds5 = ROOTDataset.from_file(['new_tree.root', 'new_chain.root'])
    ds6 = ROOTDataset.from_file('new*.root')
    del ds4, ds5, ds6
    os.remove('new_tree.root')
    os.remove('new_chain.root')

def test_selections():
    tree = create_simple_tree(200)
    sel1 = Selection("x > 1")
    sel2 = Selection("y > 1")
    sel3 = sel1 | sel2
    sel4 = sel1 & sel2
    tree_sel = tree.CopyTree(sel3.rootselection)
    assert tree_sel.GetEntries() < tree.GetEntries()
    tree_sel = tree.CopyTree(sel4.rootselection)
    assert tree_sel.GetEntries() < tree.GetEntries()

os.remove('tree1.root')
os.remove('tree2.root')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# =============================================================================
# Licensed under a 3-clause BSD style license, see LICENSE.
# =============================================================================
"""
Geometry classes
================

Three geometry classes are available:
* ``Point3D``      : a point in 3-dimensional space
* ``Line3D``       : a line  in 3-dimensional space
* ``Plane3D``      : a plane in 3-dimensional space

"""
# =============================================================================
__all__  = (
    'Vector3D' , ## 3D-vector
    'Point3D'  , ## 3D-point
    'Line3D'   , ## line in 3D
    'Plane3D'  , ## plane in 3D
    'distance' , ## 3D-distance between geometry objects
    )
# =============================================================================
# Import statements
# =============================================================================
from   skhep.math.vectors import Vector3D
from   skhep.math.numeric import isequal

# =============================================================================
# 3D-point
# =============================================================================
class Point3D(object) :
    """Point in 3D space.
    """
    def __init__ ( self , *args , **kwargs ) :
        self._vct = Vector3D( *args , **kwargs )

    @classmethod
    def origin(cls):
        """Shortcut constuctor for the origin (x=0.,y=0.,z=0.).
        Equivalent to the default constructor Point3D().
        """
        return cls(0., 0., 0.)

    @classmethod
    def frompoint(cls, x, y, z):
        """Constructor from an explicit space point."""
        return cls(x, y, z)

    @classmethod
    def fromvector(cls, other):
        """Copy constructor."""
        return cls(other.x, other.y, other.z)

    @classmethod
    def fromsphericalcoords(cls, r, theta, phi):
        """Constructor from a space point specified in spherical coordinates.

        Parameters
        ----------
        r     : radius, the radial distance from the origin (r > 0)
        theta : inclination in radians (theta in [0, pi] rad)
        phi   : azimuthal angle in radians (phi in [0, 2pi) rad)
        """
        from math import sin, cos
        x = r * sin(theta) * cos(phi)
        y = r * sin(theta) * sin(phi)
        z = r * cos(theta)
        return cls(x, y, z)

    @classmethod
    def fromcylindricalcoords(cls, rho, phi, z):
        """Constructor from a space point specified in cylindrical coordinates.

        Parameters
        ----------
        rho : radial distance from the z-axis (rho > 0)
        phi : azimuthal angle in radians (phi in [-pi, pi) rad)
        z   : height
        """
        from math import sin, cos
        x = rho * cos(phi)
        y = rho * sin(phi)
        z = z
        return cls(x, y, z)

    @classmethod
    def fromiterable(cls, values):
        """Constructor from a suitable iterable object.
        Suitable means here that all entries are numbers
        and the length equals 3.
        """
        values = list(values)
        if not len(values)==3:
            raise ValueError( 'Input iterable length = {0}! Please check your inputs.'.format(len(values)) )
        for i, v in enumerate(values):
            if not isinstance( v, (int,float) ):
                raise ValueError( 'Component #{0} is not a number!'.format(i) )
        return cls(values[0], values[1], values[2])


    @property
    def x(self):
        """Return the x, aka first coordinate at position 0."""
        return self._vct.x

    @x.setter
    def x(self, value):
        """Sets x, aka first coordinate at position 0."""
        self._vct.x = value

    @property
    def y(self):
        """Return the y, aka second coordinate at position 1."""
        return self._vct.y

    @y.setter
    def y(self, value):
        """Sets y, aka second coordinate at position 1."""
        self._vct.y = value

    @property
    def z(self):
        """Return the z, aka third coordinate at position 2."""
        return self._vct.z

    @z.setter
    def z(self, value):
        """Sets z, aka third coordinate at position 2."""
        self._vct.z = value

    @property
    def rho(self):
        """Return the cylindrical coordinate rho."""
        return self._vct.rho

    @property
    def theta(self, deg=False):
        """Return the spherical coordinate theta.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self._vct.theta( deg )

    @property
    def phi(self, deg=False):
        """Return the spherical or cylindrical coordinate phi.

        Parameters
        ----------
        deg : float, optional
            Return the angle in degrees (default is radians).
        """
        return self._vct.phi( deg )

    def copy  ( self ) :
        """Make a copy of this point."""
        return Point3D( self[0], self[1],  self[2] )

    def __setitem__(self, i, value) :
       """Update/set the ith points component (commencing at 0, of course)."""
       self._vct[i] = value

    def __getitem__(self, i):
        """Get the ith point component (commencing at 0, of course)."""
        return self._vct[i]

    def tolist(self):
        """Return the vector as a list."""
        return self._vct.tolist()

    def __len__(self):
        """Length/dimension of the poit, i.e. the number of elements = 3."""
        return 3

    @property
    def mag(self):
        """Magnitude, a.k.a. norm, of the point."""
        return self._vct.mag

    @property
    def mag2(self):
        """Square of the magnitude, a.k.a. norm, of the point."""
        return self._vct.mag2

    def __eq__  ( self , another ) :
        """Equality of two points.

        Example
        -------
        >>> p1 = ...
        >>> p2 = ...
        >>> print  p1 == p2
        """
        return self._vct ==  another._vct

    def __ne__  ( self , another ) :
        """Non-equality of two points.

        Example
        -------
        >>> p1 = ...
        >>> p2 = ...
        >>> print  p1 != p2
        """
        return self._vct !=  another._vct

    ## operations
    def __iadd__ ( self , vector ) :
        """Add vector to point.

        Example
        -------
        >>> p  = Point3D  ( ... )
        >>> v  = Vector3D ( ... )
        >>> p += v
        """

        if isinstance ( vector , Vector3D ) :
            self._vct += vector
            return self
        return NotImplemented

    def __isub__ ( self , vector ) :
        """Subtract vector from point.

        Example
        -------
        >>> p  = Point3D  ( ... )
        >>> v  = Vector3D ( ... )
        >>> p -= v
        """

        if isinstance ( vector , Vector3D ) :
            self._vct -= vector
            return self
        return NotImplemented

    ## operations
    def __add__ ( self , vector ) :
        """Addition of point and vector.

        point + vector  --> point
        """
        ## point + vector = point
        if isinstance ( vector , Vector3D ) :
            newpoint  = self.copy()
            newpoint += vector
            return newpoint

        return NotImplemented

    def __sub__ (  self , other ) :
        """ Subtraction for points and vectors.

        1. point - point   -->  vector
        2. point - vector  -->  point
        """
        ## point - point  --> vector
        if isinstance ( other , Point3D ) : return self._vct - other._vct

        ## point - vector --> point
        if isinstance ( other , Vector3D ) :
            newpoint  = self.copy()
            newpoint -= other
            return newpoint

        return NotImplemented

    def __radd__ ( self , other )  :
        """Right addition of vector and point."""
        return self + other

    def __repr__(self):
        """Class representation."""
        return "<Point3D(x={0},y={1},z={2})>".format(self.x,self.y,self.z)

    def __str__(self):
        """Simple class representation."""
        return "Point3D(x={0},y={1},z={2})".format(self.x,self.y,self.z)

    def __eq__ ( self , other ) :
        """Equality  criteria for two points.

        Example
        -------
        >>> point1 = ...
        >>> point2 = ...
        >>> point1 == point2
        """
        if   isinstance ( other , Point3D ) :
            return self._vct == other._vct
        elif isinstance ( other , ( float , int , long) ) :
            return self._vct == other
        return NotImplemented

    def __ne__ ( self , other ) :
        """Non-equality  crietria for two points.

        Example
        -------
        >>> point1 = ...
        >>> point2 = ...
        >>> point1 != point2
        """
        return not  ( self == other )

    def on_line  ( self , line ) :
        """In the point on the line?"""
        v = line.point - self
        return 0 == v.mag2 or line.iscollinear  ( v )

    def on_plane ( self , plane ) :
        """Is the point on the plane?"""
        v = plane.point - self
        return 0 == v.mag2 or plane.normal.isperpendicular ( v )

    def distance ( self , other ) :
        """Get 'distance' between point  and other object.
        """
        if   isinstance ( other , Point3D ) :
            return abs ( self - other )
        elif isinstance ( other ,  ( Line3D , Plane3D ) ) :
            return other.distance ( self )
        raise NotImplementedError("Distance from Point3D to %s is not defined" % other )


# =============================================================================
# 3D-Line
# =============================================================================
class Line3D(object) :
    """Line in 3D space.
    Line is defined by a point on line and by the direction.
    """
    def __init__ ( self                            ,
                   point  = Point3D  ( 0  ,0 , 0 ) ,
                   vector = Vector3D ( 0 , 0 , 1 ) ) :

        if not isinstance ( point  ,  Point3D  ) :
            raise NotImplementedError("Line3D: invalid ``point'' argument" )
        if not isinstance ( vector ,  Vector3D ) :
            raise NotImplementedError("Line3D: invalid ``vector'' argument" )

        if 0 == vector.mag2  :
            raise ValueError ("Line3D: ``vector'' must not be zero!" )

        self.point     = point  .copy()
        self.direction = vector .copy()

    @classmethod
    def from_points ( cls , point1 , point2 ) :
        """Create the line from two points.

        Parameters
        ----------
        point1 : the first point on the line
        point2 : the second point on the line

        """
        return cls ( point1 , point2 - point1 )

    @classmethod
    def from_line   ( cls , line  ) :
        """Create the line from anoother line."""
        return cls ( line.point , line.direction )

    def __repr__(self):
        """Simple class representation."""
        return "<Line3D({0},{1})>".format( self.point , self.direction )

    def __str__(self):
        """Simple class representation."""
        return "Line3D({0},{1})".format( self.point , self.direction )

    def copy  ( self ) :
        """Make a copy of this line."""
        return Line3D( self.point.copy() , self.direction.copy() )

    ## is the point on line ?
    def __contains__ ( self , point ) :
        """Is the point on the line?

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> if point in line : ...
        """
        return point.on_line ( self )

    def iscollinear ( self , other ) :
        """Is the line collinear to another line or vector?
        """
        if   isinstance ( other , Line3D   ) : other = other.direction
        elif isinstance ( other , Vector3D ) : pass
        else : raise TypeError("Line3D.iscollinear: invalid ``other'' argument" )
        ##
        return self.direction.iscollinear ( other  )

    def __eq__ ( self , other ) :
        """Equality  criteria for two lines.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> print line1 == line2
        """
        return other.point in self  and self.iscollinear ( other )

    def __ne__ ( self , other ) :
        """Nonequality  criteria for two lines.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> print line1 != line2
        """
        return not ( self  == other )

    def on_plane ( self , plane ) :
        """Is the line on the plane?
        """
        return self.point.on_plane ( plane ) and self.direction.isperpendicular ( plane.normal )

    def point_on_line (  self , mu ) :
        """ Get the point on line, that  corresponds to parameter 'mu'.

        Example
        -------
        >>> line = ...
        >>> p0 = line.point_on_line ( 0 ) ## should be equal to line.point
        >>> p1 = line.point_on_line ( 1 ) ## shodul be equal to line.point+line.vector
        """
        return self.point + mu * self.direction

    def closest_point_param ( self , point) :
        """Get the 'mu' parameter of the point on line that is closest to the given point in space.

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> mu    = line.closest_point_param ( point )
        """
        return ( self.direction * ( self.point - point ) ) / self.direction.mag2

    def closest_points_params ( self , line ) :
        """Get the 'mu' parameters of two points on lines that is closest to each other.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> mu1 , mu2  = line1.closest_points_param ( line2 )
        """
        a00  =  self.direction.mag2
        a01  =  self.direction * line.direction
        a10  = -a01
        a11  = -line.direction.mag2

        deti = a00 * a11 -  a01 * a10


        ## two lines are actually parallel
        if isequal ( deti , 0 , scale = abs ( a00 * a11) ) :
            return 0., line.closest_point_param ( self.point )

        detinv = 1.0/deti

        v  = line.point - self.point

        b0 = v * self.direction
        b1 = v * line.direction

        mu1 = (   a11 * b0 - a10 * b1 ) * detinv
        mu2 = (  -a01 * b0 + a00 * b1 ) * detinv

        return mu1 , mu2

    def closest_point  ( self , point ) :
        """Get the point on line that is closest to the given point in space.

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> closest  = line.closest_point ( point )
        """
        return self.point_on_line ( self.closest_point_param ( point ) )

    def closest_points ( self , line ) :
        """Get two points on lines that is closest to each other.

        Example
        -------
        >>> line1 = ...
        >>> line2 = ...
        >>> p1 , p2  = line1.closest_points ( line2 )
        """
        mu1,mu2 = self.closest_points_params ( line )
        return self.point_on_line ( mu1 ), line.point_on_line ( mu2 )

    def distance ( self , other ) :
        """Get 'distance' between line and other object.

        Examples
        --------
        >>> line  = ...
        >>> point = ...
        >>> line2 = ...
        >>> plane = ...
        >>> print line.distance ( line2 )
        >>> print line.distance ( point )
        >>> print line.distance ( plane )
        """
        if   isinstance ( other , Point3D ) :
            if other in self : return 0
            return abs ( other -  self.closest_point ( other ) )
        elif isinstance ( other ,  Line3D ) :
            if other ==  self : return 0
            p1,p2 = self.closest_points ( other )
            return abs ( p1 - p2 )
        elif isinstance ( other , Plane3D ) :
            return other.distance ( self )

        raise NotImplementedError("Distance from Line3D to %s is not defined" % other )

    def angle ( self, other ) :
        """Get the 'angle' between a line and other object.
        """
        if   isinstance ( other ,  Line3D ) :
            return self.direction.angle ( other.direction )
        elif isinstance ( other , Plane3D ) :
            return self.direction.angle ( other.normal )

        raise NotImplementedError("Angle between a Line3D and a %s is not defined" % other )

    def intersect ( self, other ) :
        """Get the 'intersection' between a line and other object.
        """

        if isinstance ( other , Point3D ):
            if other in self:
                return other
            else:
                return None

        elif isinstance ( other , Line3D ):
            dist = self.distance ( other )

            if isequal ( dist, 0.0 ):
                return self.closest_points( other )[0]
            else:
                return None

        elif isinstance ( other , Plane3D ):

            if self.direction.isperpendicular ( other.normal):
                if self.on_plane ( other ):
                    return self
                else:
                    return None
            else:
                nom = other.normal * ( other.point - self.point )
                den = other.normal * ( self.direction )
                mu = nom / den

                return self.point_on_line ( mu )

        raise NotImplementedError("Intersection between a Line3D and a %s is not defined" % other )



# =============================================================================
# 3D-Plane
# =============================================================================
class Plane3D(object) :
    """Plane in 3D space, defined by a point on plane and by the normal vector.
    """
    def __init__ ( self                            ,
                   point  = Point3D  ( 0  ,0 , 0 ) ,
                   normal = Vector3D ( 0 , 0 , 1 ) ) :

        if not isinstance ( point  ,  Point3D  ) :
            raise NotImplementedError("Plane3D: invalid ``point'' argument" )
        if not isinstance ( normal ,  Vector3D ) :
            raise NotImplementedError("Plane3D: invalid ``normal'' argument" )

        if 0 == normal.mag2  :
            raise ValueError ("Plane3D: ``normal'' must not be zero!" )

        self.point     = point  .copy()
        self.normal    = normal .copy()

    @classmethod
    def from_points ( cls , point1 , point2 , point3 ) :
        """Create the plane from three  points.

        Parameters
        ----------
        point1 : the first  point on the plane
        point2 : the second point on the plane
        point3 : the third  point on the plane
        """
        v21 = point2 - point1
        v31 = point3 - point1
        return cls ( point1 , v21.cross ( v31 ) )

    @classmethod
    def from_line_and_point ( cls , line , point ) :
        """Create the plane from the line and point.

        Parameters
        ----------
        line   : the line in plane
        point  : the point in plane
        """
        return cls ( point ,  line.direction.cross ( point - line.point ) )

    ## is the point or line on plane ?
    def __contains__ ( self , obj ) :
        """Is the point or line on the line?

        Parameters
        ----------
        object : the point or line to be tested

        Example
        -------
        >>> line  = ...
        >>> point = ...
        >>> plane = ...
        >>> if point in plane : ...
        >>> if line  in plane : ...
        """
        return obj.on_plane ( self )

    def isparallel( self , other ) :
        """Is the plane collinear to another line, vector or plane?
        """
        if   isinstance ( other , Line3D   ) :
            return self.normal.isperpendicular ( other.direction )
        elif isinstance ( other , Vector3D ) :
            return self.normal.isperpendicular ( other )
        elif isinstance ( other , Plane3D ) :
            return self.normal.iscollinear ( other.normal )
        else : raise TypeError("Plane3D.isparallel: invalid ``other'' argument" )

    def __repr__(self):
        """Simple class representation."""
        return "<Plane3D({0},{1})>".format( self.point , self.normal )

    def __str__(self):
        """Simple class representation."""
        return "Plane3D({0},{1})".format( self.point , self.normal )

    def copy  ( self ) :
        """Make a copy of this plane.
        """
        return Plane3D( self.point.copy() , self.normal.copy() )


    def __eq__ ( self , other ) :
        """Equality  criteria for two planes.

        Example
        -------
        >>> plane1 = ...
        >>> plane2 = ...
        >>> print plane1 == plane2
        """
        return other.point in self  and self.normal.iscollinear ( other.normal )

    def __ne__ ( self , other ) :
        """Nonequality  criteria for two planes.

        Example
        -------
        >>> plane1 = ...
        >>> plane2 = ...
        >>> print plane1 != plane2
        """
        return not ( self  == other )

    def distance ( self , other ) :
        """Get 'distance' between plane and other object.

        Examples
        --------
        >>> line  = ...
        >>> point = ...
        >>> line2 = ...
        >>> plane = ...
        >>> print line.distance ( line2 )
        >>> print line.distance ( point )
        >>> print line.distance ( plane )
        """
        try:
            if   isinstance ( other , Point3D ) :
                if other in self : return 0
                v = other - self.point
                return   abs( v * self.normal ) / self.normal.mag

            elif self.isparallel ( other ) :
                return self.distance ( other.point )

            else : return 0

        except TypeError:
            raise NotImplementedError("Distance from Line3D to %s is not defined" % other )


    def angle ( self, other ) :
        """Get the 'angle' between a plane and other object.
        """
        if   isinstance ( other ,  Line3D ) :
            return self.normal.angle ( other.direction )
        elif isinstance ( other , Plane3D ) :
            return self.normal.angle ( other.normal )

        raise NotImplementedError("Angle between a Line3D and a %s is not defined" % other )

    def intersect ( self, other ) :
        """Get the 'intersection' between a line and other object
        """

        if isinstance ( other , Point3D ):
            if other in self:
                return other
            else:
                return None

        elif isinstance ( other , Line3D ):
            return other.intersect ( self )

        elif isinstance ( other , Plane3D ):

            if self.isparallel( other ):
                return None
            else:

                N1, N2 = self.normal, other.normal
                d1, d2 = N1 * self.point._vct, N2 * other.point._vct

                det = N1 * N1 * N2 * N2 - ( N1 * N2 )**2
                deti = 1.0 / det

                c1 = ( d1 * N2 * N2 - d2 * N1 * N1 ) * deti
                c2 = ( d1 * N2 * N2 - d2 * N1 * N1 ) * deti

                point = Point3D.fromvector ( c1 * N1 + c2 * N2 )
                direction = N1.cross( N2 )

                return Line3D( point, direction )

        raise NotImplementedError("Intersection between a Plane3D and a %s is not defined" % other )



# ===========================================================================
def distance ( obj1 , obj2 ) :
    """Calculate 3D-distance between two geometry objects (points, lines, planes).

    Examples
    --------
    >>> line1  = ...
    >>> line2  = ...
    >>> point1 = ...
    >>> point2 = ...
    >>> plane1 = ...
    >>> plane2 = ...
    >>> for o1 in ( line1, line2 , point1 , point2 , plane1 , plane2 ) :
    ...    for o2 in ( line1, line2 , point1 , point2 , plane1 , plane2 ) :
    ...        print distance ( o1 , o2 )
    """
    return obj1.distance ( obj2 )

# ===========================================================================
if '__main__' == __name__ :

    v = Vector3D( 1,2,3)
    p = Point3D( 1,2,3)
    p += 0.5*v
    p -= 0.5*v
    p, p + v*4, p - ( p + v )

    line = Line3D ( p , v )
    p in line

    line = Line3D.from_points ( p , Point3D(3,2,1) )
    p in line

    plane = Plane3D( p , v )
    line in plane

    plane = Plane3D.from_points ( p , Point3D(3,2,1) , Point3D(0,0,3) )
    p in plane



import os
import sys
import matplotlib as mpl

if os.environ.get('DISPLAY', '') == '':
    print('no display found. Using non-interactive Agg backend')
    mpl.use('Agg')

import numpy as np
import matplotlib.pyplot as plt
import skhep
from skhep.visual import MplPlotter as skh_plt
import pytest

answer_dir = os.path.dirname(skhep.__file__)+'/../tests/data/visual'


def test_plot_object():
    obj = skh_plt(show_backend=True)
    assert(not (isinstance(type(obj), type(None))))


def test_simple_hist1(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0])

    if cmdopt == "generate":
        with open(answer_dir+'/answers_simple_hist1.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_simple_hist1')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_simple_hist1.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_simple_hist2(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], weights=data_gen[2], bins=20, normed=True,
                          scale='binwidth', color='red', histtype='step')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_simple_hist2.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_simple_hist2')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_simple_hist2.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_simple_hist3(cmdopt, data_gen):
    output = skh_plt.hist(data_gen[0], weights=data_gen[2], bins=range(5), normed=True,
                          scale='binwidth', color='red', histtype='step')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_simple_hist3.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_simple_hist3')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_simple_hist3.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_simple_hist4(cmdopt, data_gen):
    output = skh_plt.hist(data_gen[0], weights=data_gen[2], bins=range(5), normed=True,
                          scale='binwidth', color='red', histtype='bar')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_simple_hist4.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_simple_hist4')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_simple_hist4.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_blocks_hist(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], bins='blocks', scale='binwidth', color='green')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_blocks_hist.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_blocks_hist')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_blocks_hist.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_blocks_hist2(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], weights=data_gen[2], bins='blocks', scale='binwidth',
                          color='green', p0=0.1)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_blocks_hist2.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_blocks_hist2')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_blocks_hist2.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))

    with pytest.raises(ValueError):
        skh_plt.hist([data_gen[0], data_gen[1]], bins='blocks', scale='binwidth', color='green',
                     gamma=0.1)


def test_multi_hist1(cmdopt, data_gen):

    output = skh_plt.hist([data_gen[0], data_gen[1]], bins=20, normed=True)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_multi_hist1.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1])
        plt.title('test_multi_hist1')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_multi_hist1.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))


def test_error_bars(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], bins=20, errorbars=True, err_return=True, scale=5)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars2(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], bins=1, errorbars=True, scale=0.5, normed=True,
                          err_color='k', alpha=0.1, err_type='poisson', err_return=True)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars2.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars2')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars2.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars3(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], bins=20, errorbars=range(20), err_return=True,
                          histtype='step')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars3.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars3')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars3.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars4(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], bins=50, errorbars=True, err_return=True,
                          histtype='step', err_type='poisson', suppress_zero=True, scale='binwidth')

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars4.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars4')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars4.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars_weighted(cmdopt, data_gen):

    output = skh_plt.hist(data_gen[0], weights=data_gen[2], bins=20, errorbars=True,
                          err_return=True)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars_weighted.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars_weighted')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars_weighted.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars_multi(cmdopt, data_gen):

    output = skh_plt.hist([data_gen[0], data_gen[1]], bins=20, stacked=False, errorbars=True,
                          err_return=True,)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars_multi.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_multi')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars_multi.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars_stacked(cmdopt, data_gen):

    output = skh_plt.hist([data_gen[0], data_gen[1]], bins=20, stacked=True, errorbars=True,
                          err_return=True)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars_stacked.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars_stacked')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars_stacked.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars_stacked2(cmdopt, data_gen):

    output = skh_plt.hist([data_gen[0], data_gen[1]], bins=20, histtype='barstacked',
                          weights=[data_gen[2], data_gen[2]], errorbars=True, err_return=True,
                          normed=True, scale=2)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars_stacked2.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars_stacked2')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars_stacked2.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_error_bars_stacked3(cmdopt, data_gen):

    output = skh_plt.hist([data_gen[0], data_gen[1]], bins=20, histtype='step', stacked=True,
                          weights=[data_gen[2], data_gen[2]], errorbars=True, err_return=True,
                          normed=True, scale=2)

    if cmdopt == "generate":
        with open(answer_dir+'/answers_error_bars_stacked3.npz', 'wb') as f:
            np.savez(f, bc=output[0], be=output[1], berr=output[2])
        plt.title('test_error_bars_stacked2')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_error_bars_stacked3.npz')
        assert(np.all(output[0] == answers['bc']))
        assert(np.all(output[1] == answers['be']))
        assert(np.all(output[2] == answers['berr']))


def test_hist_fails(cmdopt, data_gen):
    with pytest.raises(ValueError):
        skh_plt.hist([data_gen[0], data_gen[1]], stacked=True, histtype='marker')
    with pytest.raises(ValueError):
        skh_plt.hist([data_gen[0], data_gen[1]], histtype='marker')
    with pytest.raises(KeyError):
        skh_plt.hist(1, err_return=True)
    with pytest.raises(ValueError):
        skh_plt.hist([data_gen[0], data_gen[1]], weights=data_gen[2])
    with pytest.raises(ValueError):
        skh_plt.hist(data_gen[0], weights=data_gen[2][0:10])
    with pytest.raises(KeyError):
        skh_plt.hist(data_gen[0], err_type='fake', errorbars=True)

    output1 = skh_plt.hist(5)
    assert(np.all(output1[0] == 1))
    output2 = skh_plt.hist([], range=(0, 1))
    assert(np.all(output2[0] == 0))


def test_ratio_plot(cmdopt, data_gen):

    output = skh_plt.ratio_plot(dict(x=data_gen[0], errorbars=True, histtype='marker'),
                                dict(x=data_gen[1], weights=data_gen[2], errorbars=True))

    if cmdopt == "generate":
        with open(answer_dir+'/answers_ratio_plot.npz', 'wb') as f:
            np.savez(f, bc1=output[1][0], be1=output[1][1],
                     bc2=output[2][0], be2=output[2][1])
        output[0][0].set_title('test_ratio_plot')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_ratio_plot.npz')
        assert(np.all(output[1][0] == answers['bc1']))
        assert(np.all(output[1][1] == answers['be1']))
        assert(np.all(output[2][0] == answers['bc2']))
        assert(np.all(output[2][1] == answers['be2']))


def test_ratio_plot_stacked(cmdopt, data_gen):

    output = skh_plt.ratio_plot(
        dict(x=[data_gen[0], data_gen[1]], stacked=True, errorbars=True),
        dict(x=[data_gen[0], data_gen[1]], weights=[data_gen[2], data_gen[2]], stacked=True,
             errorbars=True, err_style='line'),
        range=(-5, 5), bins='blocks'
    )

    if cmdopt == "generate":
        with open(answer_dir+'/answers_ratio_plot_stacked.npz', 'wb') as f:
            np.savez(f, bc1=output[1][0], be1=output[1][1],
                     bc2=output[2][0], be2=output[2][1])
        output[0][0].set_title('test_ratio_plot_stacked')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_ratio_plot_stacked.npz')
        assert(np.all(output[1][0] == answers['bc1']))
        assert(np.all(output[1][1] == answers['be1']))
        assert(np.all(output[2][0] == answers['bc2']))
        assert(np.all(output[2][1] == answers['be2']))


def test_ratio_plot_log(cmdopt, data_gen):

    output = skh_plt.ratio_plot(dict(x=data_gen[0], errorbars=True, histtype='marker', log=True,
                                     err_x=False),
                                dict(x=data_gen[1], weights=data_gen[2], errorbars=True),
                                logx=True, ratio_range=(0, 10))

    if cmdopt == "generate":
        with open(answer_dir+'/answers_ratio_plot_log.npz', 'wb') as f:
            np.savez(f, bc1=output[1][0], be1=output[1][1],
                     bc2=output[2][0], be2=output[2][1])
        output[0][0].set_title('test_ratio_plot_log')
        plt.show()
    elif cmdopt == "test":
        answers = np.load(answer_dir+'/answers_ratio_plot_log.npz')
        assert(np.all(output[1][0] == answers['bc1']))
        assert(np.all(output[1][1] == answers['be1']))
        assert(np.all(output[2][0] == answers['bc2']))
        assert(np.all(output[2][1] == answers['be2']))


def test_ratio_plot_quick(cmdopt, data_gen):
    # bin tests
    with pytest.raises(KeyError):
        skh_plt.ratio_plot(dict(x=data_gen[0], bins=10), dict(x=data_gen[1], bins=11))
    output = skh_plt.ratio_plot(dict(x=data_gen[0]), dict(x=data_gen[1], bins=11))
    assert(len(output[1][0]) == 11)
    # range tests
    with pytest.raises(KeyError):
        skh_plt.ratio_plot(dict(x=data_gen[0], range=(0, 1)), dict(x=data_gen[1], range=(1, 2)))
    output = skh_plt.ratio_plot(dict(x=data_gen[0], range=(-0.1, 0.1)), dict(x=data_gen[1]))
    assert(output[1][1][0] >= -0.1 and output[1][1][-1] <= 0.1)
    output = skh_plt.ratio_plot(dict(x=data_gen[0]), dict(x=data_gen[1], range=(-0.1, 0.1)))
    assert(output[1][1][0] >= -0.1 and output[1][1][-1] <= 0.1)



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
**********************
Module for ROOTDataset
**********************

The ``ROOTDataset`` class is the implementation of the ``Dataset`` abstract base class
for the [ROOT]_ package.

Note: usage of course requires that ROOT is installed.

**References**

.. [ROOT] https://root.cern.ch/.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from os.path import isfile
import glob
from types import MethodType

from ..utils.py23 import *
from ..utils.decorators import inheritdoc
from ..utils.dependencies import softimport
from ..utils.provenance import FileOrigin, ObjectOrigin, Transformation, Formatting

from .defs import *

ROOT = softimport("ROOT")
root_numpy = softimport("root_numpy")


# -----------------------------------------------------------------------------
# ROOTDataset
# -----------------------------------------------------------------------------
class ROOTDataset(FromFiles, ToFiles, NewNumpy, Dataset):
    def __init__(self, data, provenance=None):
        """Default constructor for ROOTDataset.

        Parameters
        ----------
        data: a TTree or TChain instance.
        provenance: history of the data before being wrapped as a ROOTDataset.
        """
        self._data = data
        if isinstance(data, ROOT.TChain) and (provenance is None):
            provenance = FileOrigin([f.GetTitle() for f in data.GetListOfFiles()])
        elif isinstance(data, ROOT.TTree) and (provenance is None):
            provenance = ObjectOrigin(data.GetName())
        self._provenance = provenance

    @property
    @inheritdoc(Dataset)
    def datashape(self):
        raise NotImplementedError  # TODO!

    @property
    @inheritdoc(Dataset)
    def immutable(self): return False

    @property
    @inheritdoc(Dataset)
    # ROOT datasets exist in a form that survives the Python session (feature that is different from Numpy)
    def persistent(self):
        return True

    @staticmethod
    def from_file(files, **options):
        """
        Load a dataset from a file or collection of files.

        files: string file name, glob pattern or iterable of string file names.
        options:
            treename: str, name of the TTree object in the collection of files.
                      Optional only if all input files contain a single TTree, with the same name.
        """
        if isinstance(files, str):
            files = glob.glob(files)
        treename = options.get("treename")
        from root_numpy import list_trees #TODO: avoid the dependency
        if treename is None:
            notOK = []
            treenames = set()
            for fname in files: #TODO: check the case when there are no trees in a file!
                trees = list_trees(fname)
                if len(trees) != 1: notOK.append(fname)
                treenames.add(trees[0])
            if notOK:
                raise ValueError("Multiple trees found in file(s) {0}!\nPlease specify a tree name.".format(','.join(notOK)))
            if len(treenames) > 1:
                raise ValueError("Different tree names found!\nPlease specify the desired tree name.")
            treename = treenames.pop()
            del notOK, treenames
        else:
            notOK = []
            for fname in files: #TODO: check the case when there are no trees in a file!
                trees = list_trees(fname)
                if trees[0] != treename:  notOK.append(fname)
            if notOK:
                raise ValueError("Specified tree name '{0}'' not found in file(s) {1}!\nPlease check your inputs.".format(treename,','.join(notOK)))

        chain = ROOT.TChain(treename)
        for name in list(files):
            n = chain.Add(name)

        return ROOTDataset(chain, FileOrigin([f.GetTitle() for f in chain.GetListOfFiles()]))

    @inheritdoc(ToFiles, gap="\n")
    def to_file(self, base, **options):
        """
        options:
            mode: ROOT's mode in which the TFile is to be opened. Default='update'.
        """
        mode = 'update' if options.get("mode") is None else options.get("mode")
        outfile = ROOT.TFile(base, mode)
        if not outfile:
            raise IOError("Cannot open file {0}! TTree not saved.".format(base))
        if not outfile.IsWritable():
            raise IOError("File {0} is not writable!".format(base))
        # If a tree with that name exists, we want to update it
        treename = self.data.GetName()
        tree = outfile.Get(treename)
        tree = self.data
        tree.Write(treename, ROOT.TObject.kOverwrite) # Save only the new version of the tree
        del tree
        outfile.Close()
        del outfile

    @inheritdoc(NewNumpy, gap="\n")
    def to_array(self, **options):
        """
        Parameters
        ----------
        options: see options of root_numpy.tree2array.

        Returns
        -------
        NumpyDataset holding the new NumPy structured array.
        """
        ar = root_numpy.tree2array(self.data, options)
        from .numpydataset import NumpyDataset
        return NumpyDataset(ar, self._provenance+(Formatting('NumpyDataset'),))

    def __getitem__(self, name):
        raise NotImplementedError  # TODO!


# -----------------------------------------------------------------------------
# Add ROOT methods to ROOTDataset in bulk.
# -----------------------------------------------------------------------------
def addROOTMethod(method):
    raise NotImplementedError  # TODO!


try:
    pass
    # ... (see numpydataset.py)
except ImportError:
    pass



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Tests for the skhep.dataset.defs module.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
import pytest

from skhep.dataset.defs import *


# -----------------------------------------------------------------------------
# Actual tests
# -----------------------------------------------------------------------------

def test_base_classes():
    with pytest.raises(TypeError):
        Dataset()

def test_mixins():
    FromPersistent()
    ToPersistent()
    ConvertibleCopy()
    ff = FromFiles()
    with pytest.raises(NotImplementedError):
        ff.from_file('non_existent_file')
    tf = ToFiles()
    with pytest.raises(NotImplementedError):
        tf.to_file('non_existent_file')
    np = NewNumpy()
    with pytest.raises(NotImplementedError):
        np.to_array()
    nr = NewROOT()
    with pytest.raises(NotImplementedError):
        nr.to_tree()



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Submodule for useful exceptions
===============================

.. note:: not meant for user code in general, though possible.
"""

# Definition of handy colours for printing
_default = '\x1b[00m'
_green   = '\x1b[01;32m'
_red     = '\x1b[01;31m'

class InvalidOperationError(Exception):
    """Exception class for meaningless operations."""
    def __init__(self,*args,**kwargs):
        Exception.__init__(self,*args,**kwargs)
    def __str__(self):
        """String representation."""
        return ( _red + self.message + _default )

class SkhepTypeError(Exception):
    """Exception class for non-instantiable classes."""
    def __init__(self,name):
        Exception.__init__(self,name)
        self.message = "'{0}' is an abstract base class. Instantiate one of its subclasses instead.".format(name)
    def __str__(self):
        """String representation."""
        return ( _red + self.message + _default )



# Licensed under a 3-clause BSD style license, see LICENSE.
"""
**************************************
Module for dataset-related definitions
**************************************

The module contains the definition of the ``Dataset`` abstract base class
and a series of mixin classes.
"""

# -----------------------------------------------------------------------------
# Import statements
# -----------------------------------------------------------------------------
from __future__ import absolute_import

import glob

from ..utils.py23 import *


# -----------------------------------------------------------------------------
# Generic Dataset
# -----------------------------------------------------------------------------
class Dataset(object):
    """
    Abstract base class for all dataset classes.
    """

    def __init__(self):
        raise TypeError(
            "Dataset is an abstract base class; instantiate one of its subclasses insteadone of its subclasses instead.")

    @property
    def data(self):
        """The actual data underlying this dataset.

        Use this to pass the data to an external library (without provenance).
        """
        return self._data

    @property
    def provenance(self):
        """The series of transformations that produced this dataset as a tuple."""
        return self._provenance

    @property
    def datashape(self):
        """Every dataset has a datashape, which describes its data types in a unified way."""
        # There is a standard for this:
        #     http://libndtypes.readthedocs.io/en/latest/datashape.html
        # We should use it!
        raise NotImplementedError

    @property
    def immutable(self):
        """If True, this dataset cannot be modified in place, only transformed.

        Opposite of mutable.
        """
        raise NotImplementedError

    @property
    def mutable(self):
        """If True, this dataset can be modified in place.

        Opposite of immutable.
        """
        return not self.immutable

    @property
    def persistent(self):
        """If True, this dataset exists in a form that survives the Python session, such as a file or database.

        If mutable, changes in this dataset are reflected in that persistent form.

        Opposite of transient.
        """
        raise NotImplementedError

    @property
    def transient(self):
        """If True, this dataset only exists in the Python session; changes will be lost if it is not saved.

        If mutable, changes in this dataset do not propagate back to its source, if any.

        Opposite of persistent.
        """
        return not self.persistent


# -----------------------------------------------------------------------------
# Mixins declaring functionality
# -----------------------------------------------------------------------------

# By adding methods with mixins, we enforce more uniformity than duck typing.
# Only supported methods appear when the user tab-completes.
# A generic dataset can be tested for functionality with isinstance.
class FromPersistent(object):
    pass


class ToPersistent(object):
    pass


# Uncomment when actually using/implementing this
#class ConvertibleInPlace(object):
#    pass


class ConvertibleCopy(object):
    pass


class FromFiles(FromPersistent):
    @staticmethod
    def _openFileNames(files):
        if isinstance(files, string_types):
            files = glob.glob(files)
        for x in files:
            if isinstance(x, string_types):
                x = open(x, "rb")
            assert isinstance(x, file) and "r" in x.mode
            yield x

    @staticmethod
    def from_file(files, **options):
        """
        Load a dataset from a file or collection of files.

        files: a string file name (glob pattern), iterable of string file names, or an iterable of files open for reading (binary).
        """
        # NOTE: can't @inheritdoc because this is a @staticmethod
        raise NotImplementedError


class ToFiles(ToPersistent):
    @staticmethod
    def _openRolloverFiles(base, rollover_pattern=None):
        """Generic method to generate an infinite series of files with _1, _2, etc. in their names."""
        if rollover_pattern is None:
            # TODO: Rewrite this using os.path
            def rollover_pattern(base_pattern, number):
                if "." in base_pattern:
                    return (base_pattern[:base_pattern.rindex(".")] + "_" +
                            str(number) + base_pattern[base_pattern.rindex("."):])
                else:
                    return base_pattern + "_" + str(number)

        if isinstance(base, file):
            f = base
            base = f.name
        elif isinstance(base, string_types):
            f = None
        else:
            assert False, "base should be an open file or string filename."
        n = 0
        while True:
            if f is None:
                name = rollover_pattern(base, n)
                f = open(name, "wb")
            assert "w" in f.mode
            yield f   # use the first file if given, otherwise start rollover with _0
            f = None  # second file in rollover is always _1 and continuing from there

    @staticmethod
    def _openSingleFile(base):
        if isinstance(base, string_types):
            base = open(base, "wb")
        assert isinstance(base, file) and "w" in base.mode
        return base

    def to_file(self, base, **options):
        """
        Save this dataset to a file or collection of files.

        Parameters
        ----------
        base: str or iterable of str
            String file name or iterable of string file names.
        """
        raise NotImplementedError


# Uncomment when actually using/implementing this
#class AsNumpy(ConvertibleInPlace):
#    def as_array(self, **options):
#        """View this dataset as a NumpyDataset, sharing its underlying data.
#
#        A change in the NumpyDataset modifies the original.
#        """
#        raise NotImplementedError


class NewNumpy(ConvertibleCopy):
    def to_array(self, **options):
        """
        Copy this dataset into a new NumpyDataset, without sharing any underlying data.

        A change in the NumpyDataset leaves the original untouched.
        """
        raise NotImplementedError


# Uncomment when actually using/implementing this
#class AsROOT(ConvertibleInPlace):
#    def as_tree(self, **options):
#        """View this dataset as a ROOTDataset, sharing its underlying data.
#
#        A change in the ROOTDataset modifies the original.
#        """
#        raise NotImplementedError


class NewROOT(ConvertibleCopy):
    def to_tree(self, **options):
        """
        Copy this dataset into a new ROOTDataset, without sharing any underlying data.

        A change in the ROOTDataset leaves the original untouched.
        """
        raise NotImplementedError



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license, see LICENSE.
import sys
import os

# try to use setuptools if installed
from setuptools import setup, find_packages
from setuputils import read, find_version


# General information on the project
PROJECT_NAME = 'scikit-hep'
AUTHOR = 'the scikit-hep developers'
AUTHOR_EMAIL = 'scikit-hep-admins@googlegroups.com'
DESCRIPTION = 'Toolset of interfaces and tools for Particle Physics.'
URL = 'https://github.com/scikit-hep/scikit-hep/'
LICENSE = 'new BSD'
VERSION = find_version('skhep/__init__.py')

# Specification of minimal versions of required dependencies
PYTHON_REQUIRES = '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4'
NUMPY_MIN_VERSION = '1.11.0'

# Prevent setup from trying to create hard links
# which are not allowed on AFS between directories.
# This is a hack to force copying.
try:
    del os.link
except AttributeError:
    pass

LOCAL_PATH = os.path.dirname(os.path.abspath(__file__))
# setup.py can be called from outside the scikit-hep directory
os.chdir(LOCAL_PATH)
sys.path.insert(0, LOCAL_PATH)

INSTALL_REQUIRES = [
    'hepunits',
    'numpy>={0};python_version>="2.7"'.format(NUMPY_MIN_VERSION),
    'pandas<0.21;python_version=="3.4.*"',
    'pandas;python_version>="2.7"',
    'matplotlib>2.0.0;python_version>="2.7"',
]

TESTS_REQUIRE = [
    'pytest>3.0;python_version>="2.7"'
]

SETUP_REQUIRES = [
    'pytest-runner'
]

setup(
    name=PROJECT_NAME,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    version=VERSION,
    description=DESCRIPTION,
    long_description=read('README.rst'),
    url=URL,
    license=LICENSE,
    packages=find_packages(),
    py_modules=['setuputils'],
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    setup_requires=SETUP_REQUIRES,
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'License :: OSI Approved',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: C++',
        'Programming Language :: Cython',
        'Development Status :: 4 - Beta',
    ]
)
